from docutils import nodes
from docutils.parsers.rst import Directive, directives

try:
    import micawber
except ImportError:
    micawber = None  # NOQA


class Embed(Directive):
    """ Restructured text extension for inserting any
        sort of media using micawber."""
    has_content = False
    required_arguments = 1
    optional_arguments = 999

    def get_providers(self):
        return micawber.bootstrap_noembed()

    def run(self):
        if micawber is None:
            msg = "To use the embed directive, install micawber first."
            return [nodes.raw('', '<div class="text-error">{0}</div>'.format(msg), format='html')]
        url = " ".join(self.arguments)
        html = micawber.parse_text(url, self.get_providers())
        return [nodes.raw('', html, format='html')]


def register_directive():
    directives.register_directive('embed', Embed)



from functools import wraps
from collections import Iterable
from django.conf import settings
from django.shortcuts import render
from django.core.exceptions import PermissionDenied
from django.utils.decorators import available_attrs
from django.utils.encoding import force_str
from django.utils.six.moves.urllib.parse import urlparse
from django.utils.six import string_types
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.shortcuts import resolve_url
from .models import ACLRule
from .settings import (WALIKI_ANONYMOUS_USER_PERMISSIONS,
                       WALIKI_LOGGED_USER_PERMISSIONS,
                       WALIKI_RENDER_403)


def check_perms(perms, user, slug, raise_exception=False):
    """a helper user to check if a user has the permissions
    for a given slug"""
    if isinstance(perms, string_types):
        perms = {perms}
    else:
        perms = set(perms)

    allowed_users = ACLRule.get_users_for(perms, slug)

    if allowed_users:
        return user in allowed_users

    if perms.issubset(set(WALIKI_ANONYMOUS_USER_PERMISSIONS)):
        return True

    if user.is_authenticated() and perms.issubset(set(WALIKI_LOGGED_USER_PERMISSIONS)):
        return True

    # First check if the user has the permission (even anon users)
    if user.has_perms(perms):
        return True
    # In case the 403 handler should be called raise the exception
    if raise_exception:
        raise PermissionDenied
    # As the last resort, show the login form
    return False


def permission_required(perms, login_url=None, raise_exception=False, redirect_field_name=REDIRECT_FIELD_NAME):
    """
    this is analog to django's builtin ``permission_required`` decorator, but
    improved to check per slug ACLRules and default permissions for
    anonymous and logged in users

    if there is a rule affecting a slug, the user needs to be part of the
    rule's allowed users. If there isn't a matching rule, defaults permissions
    apply.
    """

    def decorator(view_func):
        @wraps(view_func, assigned=available_attrs(view_func))
        def _wrapped_view(request, *args, **kwargs):

            if check_perms(perms, request.user, kwargs['slug'], raise_exception=raise_exception):
                return view_func(request, *args, **kwargs)
            if request.user.is_authenticated():
                if WALIKI_RENDER_403:
                    return render(request, 'waliki/403.html', kwargs, status=403)
                else:
                    raise PermissionDenied

            path = request.build_absolute_uri()
            # urlparse chokes on lazy objects in Python 3, force to str
            resolved_login_url = force_str(
                resolve_url(login_url or settings.LOGIN_URL))
            # If the login url is the same scheme and net location then just
            # use the path as the "next" url.
            login_scheme, login_netloc = urlparse(resolved_login_url)[:2]
            current_scheme, current_netloc = urlparse(path)[:2]
            if ((not login_scheme or login_scheme == current_scheme) and
                    (not login_netloc or login_netloc == current_netloc)):
                path = request.get_full_path()
            from django.contrib.auth.views import redirect_to_login
            return redirect_to_login(
                path, resolved_login_url, redirect_field_name)
        return _wrapped_view

    return decorator


from django.conf.urls import patterns, url
from waliki.settings import WALIKI_SLUG_PATTERN


urlpatterns = patterns('waliki.pdf.views',
    url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/pdf$', 'pdf', name='waliki_pdf'),
)



import django.dispatch

page_preedit = django.dispatch.Signal(providing_args=["page"])
page_saved = django.dispatch.Signal(providing_args=["raw", "author", "message"])
page_moved = django.dispatch.Signal(providing_args=["page", "old_path", "author", "message"])



# -*- coding: utf-8 -*-
import os.path
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.db.models.signals import pre_delete
from django.dispatch.dispatcher import receiver
from django.utils.encoding import python_2_unicode_compatible
from django.utils.six import text_type
from django.core.urlresolvers import reverse
from waliki.models import Page
from waliki.settings import WALIKI_UPLOAD_TO


@python_2_unicode_compatible
class Attachment(models.Model):
    page = models.ForeignKey(Page, related_name='attachments')
    file = models.FileField(upload_to=WALIKI_UPLOAD_TO, max_length=300)

    class Meta:
        verbose_name = _("Attachment")
        verbose_name_plural = _("Attachments")

    def __str__(self):
    	return os.path.basename(self.file.name)

    def get_absolute_url(self):
        return reverse('waliki_attachment_file', args=(self.page.slug, self.id, text_type(self)))


# @receiver(pre_delete, sender=Attachment)
# def attachment_delete(sender, instance, **kwargs):
#    try:
#        # instance.file.delete(False)
#        pass
#    except SuspiciousFileOperation:
#        pass


import os
import re
import json
from django.contrib.auth.models import User
from django.utils.translation import ugettext_lazy as _
from django.utils import six
from waliki.models import Page
from waliki import settings
from sh import git, ErrorReturnCode, Command


git = git.bake("--no-pager")


class Git(object):
    __shared_state = {}     # it's a Borg

    def __init__(self):
        self.__dict__ = self.__shared_state
        from waliki.settings import WALIKI_DATA_DIR
        self.content_dir = WALIKI_DATA_DIR
        os.chdir(self.content_dir)
        if not os.path.isdir(os.path.join(self.content_dir, '.git')):
            git.init()
            git.config("user.email", settings.WALIKI_COMMITTER_EMAIL)
            git.config("user.name", settings.WALIKI_COMMITTER_NAME)

        self.git = git

    def commit(self, page, message='', author=None, parent=None, extra_path=None):
        path = page.path
        paths_to_commit = [path]
        if extra_path:
            paths_to_commit.append(extra_path)
        kwargs = {}
        if isinstance(author, User) and author.is_authenticated():
            kwargs['author'] = u"%s <%s>" % (author.get_full_name() or author.username, author.email)
        elif isinstance(author, six.string_types):
            kwargs['author'] = author

        try:
            there_were_changes = parent and parent != self.last_version(page)
            status = git.status('--porcelain', path).stdout.decode('utf8')[:2]
            if parent and status != "UU":
                git.stash()
                git.checkout('--detach', parent)
                try:
                    git.stash('pop')
                except:
                    git.checkout('--theirs', path)

            if status == 'UU':
                # See http://stackoverflow.com/a/8062976/811740
                kwargs['i'] = True

            git.add(path)
            git_commit_cmd = git.commit.bake(allow_empty=True, allow_empty_message=True, m=message, **kwargs)
            git_commit_cmd('--', *paths_to_commit)
            last = self.last_version(page)
            if parent and status != "UU":
                git.checkout('master')
                git.merge(last)
        except ErrorReturnCode as e:
            # TODO: make this more robust!
            error = e.stdout.decode('utf8')
            if 'CONFLICT' in error:
                # For '-i' attribute see http://stackoverflow.com/q/5827944/811740
                git_commit_cmd = git.commit.bake(allow_empty=True, allow_empty_message=True, m=_('Merged with conflict'), i=True, **kwargs)
                git_commit_cmd('--', *paths_to_commit)
                raise Page.EditionConflict(_('Automatic merge failed. Please, fix the conflict and save the page.'))
            else:
                raise
        return there_were_changes


    def history(self, page):
        GIT_COMMIT_FIELDS = ['commit', 'author', 'date', 'date_relative', 'message']
        GIT_LOG_FORMAT = '%x1f'.join(['%h', '%an', '%ad', '%ar', '%s']) + '%x1e'
        output = git.log('--format=%s' % GIT_LOG_FORMAT, '--follow', '-z', '--shortstat', page.abspath)
        output = output.split('\n')
        history = []
        for line in output:
            if '\x1f' in line:
                log = line.strip('\x1e\x00').split('\x1f')
                history.append(dict(zip(GIT_COMMIT_FIELDS, log)))
            else:
                insertion = re.match(r'.* (\d+) insertion', line)
                deletion = re.match(r'.* (\d+) deletion', line)
                history[-1]['insertion'] = int(insertion.group(1)) if insertion else 0
                history[-1]['deletion'] = int(deletion.group(1)) if deletion else 0

        max_changes = float(max([(v['insertion'] + v['deletion']) for v in history])) or 1.0
        for v in history:
            v.update({'insertion_relative': str((v['insertion'] / max_changes) * 100),
                      'deletion_relative': str((v['deletion'] / max_changes) * 100)})
        return history

    def version(self, page, version):
        try:
            return git.show('%s:%s' % (version, page.path)).stdout.decode('utf8')
        except:
            return ''

    def last_version(self, page):
        try:
            return six.text_type(git.log("--pretty=format:%h", "-n 1", page.path))
        except ErrorReturnCode:
            return None

    def whatchanged(self, skip=0, max_count=None):
        GIT_LOG_FORMAT = '%x1f'.join(['%an', '%ae', '%h', '%s', '%ar'])
        pages = []

        args = ["--pretty=format:%s" % GIT_LOG_FORMAT, '--skip=%d' % skip]
        if max_count:
            args.append('--max-count=%d' % max_count)
        raw_log = git.whatchanged(*args).stdout.decode('utf8')
        logs = re.findall(r'((.*)\x1f(.*)\x1f(.*)\x1f(.*)\x1f(.*))?\n:.*\t(.*)', raw_log, flags=re.MULTILINE | re.UNICODE)
        for log in logs:
            if log[0]:
                log = list(log[1:])
                log[-1] = [log[-1]]
                pages.append(list(log))
            else:
                pages[-1][-1].append(log[-1])
        return pages

    def pull(self, remote):
        log = git.pull('-s', 'recursive', '-X', 'ours', remote, 'HEAD').stdout.decode('utf8')
        return log

    def diff(self, page, new, old):
        return git.diff('--no-color', new, old, '--', page.path).stdout.decode('utf8')

    def total_commits(self, to='HEAD', page=None):
        args = ['rev-list', to, '--count']
        if page:
            args += ['--', page.path]
        return git(*args).stdout.decode('utf8')[:-1]

    def mv(self, sender, page, old_path, author, message, commit=True):
        status = git.status('--porcelain', old_path).stdout.decode('utf8')[1:2]
        extra_path = ''
        if status in ('D', 'M'):
            git.rm(old_path)
            extra_path = old_path
        if commit:
            self.commit(page, author=author, message=message, extra_path=extra_path)



from django.conf.urls import patterns, url, include
from django.contrib import admin
from waliki.settings import WALIKI_SLUG_PATTERN
from .plugins import load_plugins, page_urls

admin.autodiscover()
load_plugins()


def waliki_urls():
    base = [url(r'^$', 'home', name='waliki_home'),
            url(r'^_new$', 'new', name='waliki_new'),
            url(r'^_get_slug$', 'get_slug', name='waliki_get_slug'),
            url(r'^_preview$', 'preview', name='waliki_preview')]

    for pattern in page_urls():
        base.append(url(r'^', include(pattern)))

    base += [url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/edit$', 'edit', name='waliki_edit'),
             url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/delete$', 'delete', name='waliki_delete'),
             url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/move$', 'move', name='waliki_move'),
             url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/raw$', 'detail', {'raw': True}, name='waliki_detail_raw'),
             url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')$', 'detail', name='waliki_detail'),
             ]
    return base


urlpatterns = patterns('waliki.views',
    *waliki_urls()
)



import re
from waliki.signals import page_saved
from optparse import make_option
from django.core.management.base import BaseCommand, CommandError
from waliki.models import Page
from django.utils.translation import ugettext_lazy as _
from django.utils.text import get_text_list
try:
    from waliki.attachments.models import Attachment
except ImportError:
    Attachment = None

try:
    from sh import pandoc, echo
except ImportError:
    pandoc = None


def clean_meta(rst_content):
    """remove moinmoin metada from the top of the file"""

    rst = rst_content.split('\n')
    for i, line in enumerate(rst):
        if line.startswith('#'):
            continue
        break
    return '\n'.join(rst[i:])


def delete_relative_links(rst_content):
    """remove links relatives. Waliki point them correctly implicitly"""

    return re.sub(r'^(\.\. .*: \.\./.*)\n$', '', rst_content, flags=re.MULTILINE)


def attachments(rst_content, slug):

    def rep(matchobj):
        for filename in matchobj.groups(1):
            try:
                a = Attachment.objects.filter(file__endswith=filename, page__slug=slug)[0]
            except IndexError:
                print('Cant find %s in %s' % (filename, slug))
                return None
        return '`%s <%s>`_' % (filename, a.get_absolute_url())

    return re.sub(r'`attachment:(.*)`_', rep, rst_content, flags=re.MULTILINE)


def directives(rst_content):
    for directive in re.findall(r':(\w+):`.*`', rst_content, flags=re.MULTILINE):
        rst_content += """

.. role:: {directive}
   :class: {directive}

""".format(directive=directive)
    return rst_content


def emojis(rst_content):
    # require
    emojis_map = {
        ':)': 'smile',
        ':-)': 'smile',
        ';)': 'wink',
        ';-)': 'wink',
        ':-?': 'smirk',
        ':?': 'smirk',
        ':(': 'confused',
        ':-(': 'confused',
        ':D': 'laughing',
        ':-D': 'laughing',
        ':-P': 'stuck_out_tongue_closed_eyes',
        ':P': 'stuck_out_tongue_closed_eyes',
        ":'(": 'cry',
        ":'-(": 'cry',
    }

    def replace_emoji(match):
        replacement = emojis_map.get(match.groups()[0], '')
        if replacement:
            return '|%s|' % replacement
        return ''
    result = re.sub(r'\|((?:\:|;).{1,3})\|', replace_emoji, rst_content, flags=re.MULTILINE)
    return result


def email(rst_content):
    pattern = r'`\[\[MailTo\((.*)\)\]\]`_(?:\.\.)?'
    return re.sub(pattern, r'``\1``', rst_content)


def code(rst_content):
    if not pandoc:
        return rst_content
    pattern = r'^\:\:\n\s+\.\. raw:: html\n\s+(<span class\=\"line\"\>.*?|\s+?<\/span\>)\n\s*$'

    def convert(match):
        source = match.groups()[0]
        source = '\n'.join(l.strip() for l in source.split('\n'))
        source = "<pre>%s</pre>" % source
        rst_source = pandoc(echo(source), f='html', t='rst').stdout.decode('utf8')
        # rst_source = rst_source.strip().replace('\n', '\n   ') + '\n'
        return rst_source
    result = re.sub(pattern, convert, rst_content, flags=re.DOTALL | re.MULTILINE)
    return result


class Command(BaseCommand):
    help = 'Cleanup filters for a moin2git import'

    option_list = (
        make_option('--limit-to',
                    dest='slug',
                    default='',
                    help="optional namespace"),
        make_option('--filters',
                    dest='filters',
                    default='all',
                    help="comma separated list of filter functions to apply"),
        make_option('--message',
                    dest='message',
                    default=_("RestructuredText clean up"),
                    help="log message"),
    ) + BaseCommand.option_list

    def handle(self, *args, **options):
        valid_filters = ['meta', 'links',
                         'attachments', 'directives',
                         'emojis', 'title', 'email', 'code']
        slug = options['slug']
        filters = options['filters']

        if filters == 'all':
            filters = valid_filters

        else:
            filters = [f.strip() for f in filters.split(',')]
            if not set(filters).issubset(valid_filters):
                valid = get_text_list(valid_filters, 'and')
                raise CommandError("At least one filter is unknown. Valid filters are:\n    %s" % valid)

        if slug:
            pages = Page.objects.filter(slug__startswith=slug)
        else:
            pages = Page.objects.all()

        for page in pages:
            title = None
            print('\nApplying filter/s %s to %s' % (get_text_list(filters, 'and'), page.slug))

            raw = page.raw
            if 'meta' in filters:
                raw = clean_meta(raw)
            if 'links' in filters:
                raw = delete_relative_links(raw)
            if 'attachments' in filters:
                raw = attachments(raw, page.slug)

            if 'directives' in filters:
                raw = directives(raw)

            if 'emojis' in filters:
                raw = emojis(raw)

            if 'email' in filters:
                raw = email(raw)
            if 'code' in filters:
                if not pandoc:
                    print('The filter "code" need Pandoc installed in your system. Ignoring')
                else:
                    raw = code(raw)
            if 'title' in filters and not page.title:
                title = page._get_part('get_document_title')


            if raw != page.raw or title:
                if title:
                    page.title = title
                if raw != page.raw:
                    page.raw = raw
                page.save()
                page_saved.send_robust(sender='moin',
                                       page=page,
                                       author=None,
                                       message=options['message'],
                                       form_extra_data={})
            else:
                print('Nothing changed. Ignoring update')



from django.dispatch import receiver
from django.utils.translation import ugettext_lazy as _
from waliki.signals import page_saved, page_preedit, page_moved
from . import Git


@receiver(page_saved)
def commit(sender, page, author, message, form_extra_data, **kwargs):
    parent = form_extra_data.get('parent', None)
    was_moved = kwargs.get('was_moved', False)
    there_were_changes = Git().commit(page, author=author, message=message,
                                           parent=parent)
    if there_were_changes and not was_moved:
        msg = _('There were changes in the page during your edition. Auto-merge has been applied.')
        return {'messages': {'warning': msg}}


@receiver(page_preedit)
def get_last_version(sender, page, **kwargs):
    last = Git().last_version(page)
    return {'form_extra_data': {'parent': last}}


@receiver(page_moved)
def move(sender, page, old_path, author, message, commit=True, **kwargs):
    Git().mv(sender, page, old_path, author, message, commit)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('waliki', '0003_auto_20141110_0052'),
    ]

    operations = [
        migrations.CreateModel(
            name='Redirect',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, auto_created=True, verbose_name='ID')),
                ('old_slug', models.CharField(max_length=200, unique=True)),
                ('new_slug', models.CharField(max_length=200)),
                ('status_code', models.IntegerField(default=302, max_length=3, choices=[(302, '302 Found'), (301, '301 Moved Permanently')])),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



import os
from optparse import make_option
from django.conf import settings
from django.core.management.base import BaseCommand
from django.core.files import File
from waliki.settings import WALIKI_DATA_DIR, WALIKI_UPLOAD_TO
from waliki.models import Page
if 'waliki.attachments' in settings.INSTALLED_APPS:
    from waliki.attachments.models import Attachment
else:
    Attachment = None


class Command(BaseCommand):
    help = """Syncronize pages (and attachments) between files and the database"""

    option_list = (
        make_option('--extensions',
                    dest='extensions',
                    default=".rst, .md",
                    help="Look for files with this extensions, separated by comma. Default: '.rst, .md'"),
        make_option('--ignored_dirs',
                    dest='ignored_dirs',
                    default=".git",
                    help="List of directories to ignore, separated by comman. Default: '.git'"),
    ) + BaseCommand.option_list

    def handle(self, *args, **options):
        extensions = [ext.strip() for ext in options['extensions'].split(',')]
        ignored_dirs = [d.strip() for d in options['ignored_dirs'].split(',')]
        for root, dirs, files in os.walk(WALIKI_DATA_DIR):
            [dirs.remove(d) for d in ignored_dirs if d in dirs]
            for filename in files:
                if os.path.splitext(filename)[1] not in extensions:
                    continue
                path = os.path.join(root.replace(WALIKI_DATA_DIR, ''), filename).strip('/')

                if not Page.objects.filter(path=path).exists():
                    page = Page.from_path(path)
                    self.stdout.write('Created page %s for %s' % (page.get_absolute_url(), path))

        # Deleted pages?
        for page in Page.objects.all():
            if not os.path.exists(page.abspath):
                self.stdout.write('Deleted page %s (missing %s)' % (page.get_absolute_url(), page.path))
                page.delete()

        if Attachment:
            class FakeAttachment(object):
                def __init__(self, page):
                    self.page = page

            for page in Page.objects.all():
                path = os.path.join(settings.MEDIA_ROOT, WALIKI_UPLOAD_TO(FakeAttachment(page), ''))
                if not os.path.exists(path):
                    continue
                for filename in os.listdir(path):
                    if not os.path.isfile(os.path.join(path, filename)):
                        continue
                    file = WALIKI_UPLOAD_TO(FakeAttachment(page), filename)
                    if page.attachments.filter(file=filename):
                        continue
                    attachment = Attachment.objects.create(page=page, file=file)
                    self.stdout.write('Created attachment %s for %s' % (attachment, page.slug))

            for attachment in Attachment.objects.all():
                if not os.path.exists(os.path.join(settings.MEDIA_ROOT, attachment.file.name)):
                    self.stdout.write('Missing %s from %s. Deleted attachment object' % (attachment, attachment.page.slug))
                    attachment.delete()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ACLRule',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('name', models.CharField(unique=True, max_length=200, verbose_name='Name')),
                ('slug', models.CharField(max_length=200)),
                ('as_namespace', models.BooleanField(default=False, verbose_name='As namespace')),
                ('groups', models.ManyToManyField(to='auth.Group', blank=True)),
                ('permissions', models.ManyToManyField(to='auth.Permission')),
                ('users', models.ManyToManyField(to=settings.AUTH_USER_MODEL, blank=True)),
            ],
            options={
                'verbose_name_plural': 'ACL rules',
                'verbose_name': 'ACL rule',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Page',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('title', models.CharField(max_length=200, verbose_name='Title')),
                ('slug', models.CharField(unique=True, max_length=200)),
                ('path', models.CharField(unique=True, max_length=200)),
                ('markup', models.CharField(max_length=20, default='reStructuredText', choices=[('reStructuredText', 'reStructuredText'), ('Markdown', 'Markdown')], verbose_name='Markup')),
            ],
            options={
                'permissions': (('view_page', 'Can view page'),),
                'verbose_name_plural': 'Pages',
                'verbose_name': 'Page',
            },
            bases=(models.Model,),
        ),
    ]



from django.utils.translation import ugettext_lazy as _
from waliki.plugins import BasePlugin, register


class GitPlugin(BasePlugin):

    slug = 'git'
    urls_page = ['waliki.git.urls']
    extra_page_actions = {'all': [('waliki_history', _('History'))]}
    navbar_links = (('waliki_whatchanged', _('What changed')),)

register(GitPlugin)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('waliki', '0004_redirect'),
    ]

    operations = [
        migrations.AlterField(
            model_name='page',
            name='title',
            field=models.CharField(null=True, verbose_name='Title', blank=True, max_length=200),
        ),
    ]



# -*- coding: utf-8 -*-
import os
import codecs
import shutil
import os.path
from django.db import models
from django.db.models import Q
from django.db.utils import IntegrityError
from django.conf import settings
from django.core.urlresolvers import reverse
from django.dispatch import receiver
from django.utils.six import string_types
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth.models import Permission, Group, AnonymousUser
from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.db.models.signals import post_save, pre_delete
from docutils.utils import SystemMessage
from . import _markups
from waliki.settings import (get_slug, WALIKI_DEFAULT_MARKUP,
                             WALIKI_MARKUPS_SETTINGS, WALIKI_DATA_DIR,
                             WALIKI_CACHE_TIMEOUT)


class Page(models.Model):
    MARKUP_CHOICES = [(m.name, m.name) for m in _markups.get_all_markups()]
    title = models.CharField(verbose_name=_('Title'), max_length=200, blank=True, null=True)
    slug = models.CharField(max_length=200, unique=True)
    path = models.CharField(max_length=200, unique=True)
    markup = models.CharField(verbose_name=_('Markup'), max_length=20,
                              choices=MARKUP_CHOICES, default=WALIKI_DEFAULT_MARKUP)

    class Meta:
        verbose_name = _('Page')
        verbose_name_plural = _('Pages')
        permissions = (
            ('view_page', 'Can view page'),
        )

    class EditionConflict(Exception):
        pass

    def __str__(self):
        return self.__unicode__()

    def __unicode__(self):
        return self.path

    def get_absolute_url(self):
        return reverse('waliki_detail', args=(self.slug,))

    def get_edit_url(self):
        return reverse('waliki_edit', args=(self.slug,))

    def save(self, *args, **kwargs):
        self.slug = self.slug.strip('/')
        if not self.path:
            self.path = self.slug + self.markup_.file_extensions[0]
        super(Page, self).save(*args, **kwargs)

    @classmethod
    def from_path(cls, path, markup=None):
        filename, ext = os.path.splitext(path)
        if markup and isinstance(markup, string_types):
            markup = _markups.find_markup_class_by_name(markup)
        else:
            markup = _markups.find_markup_class_by_extension(ext)
        page = Page(path=path, slug=get_slug(filename), markup=markup.name)
        page.title = page._get_part('get_document_title')

        while True:
            try:
                page.save()
                break
            except IntegrityError:
                page.slug += '-'

        return page

    @property
    def raw(self):
        filename = self.abspath
        if not os.path.exists(filename) or os.path.isdir(filename):
            return ""
        return codecs.open(filename, "r", encoding="utf-8").read()

    @raw.setter
    def raw(self, value):
        filename = os.path.join(WALIKI_DATA_DIR, self.path)
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError:
            pass
        with codecs.open(filename, "w", encoding="utf-8") as f:
            f.write(value)

    @property
    def abspath(self):
        return os.path.abspath(os.path.join(WALIKI_DATA_DIR, self.path))

    def move(self, new_path):
        try:
            os.makedirs(os.path.dirname(os.path.join(WALIKI_DATA_DIR, new_path)))
        except OSError:
            pass
        shutil.move(os.path.join(WALIKI_DATA_DIR, self.path), os.path.join(WALIKI_DATA_DIR, new_path))
        self.path = new_path

    def update_extension(self):
        filename, _ = os.path.splitext(self.path)
        markup_class = _markups.find_markup_class_by_name(self.markup)
        self.move(filename + markup_class.file_extensions[0])

    @staticmethod
    def get_markup_instance(markup):
        markup_settings = WALIKI_MARKUPS_SETTINGS.get(markup, None)
        markup_class = _markups.find_markup_class_by_name(markup)
        return markup_class(**markup_settings)

    @staticmethod
    def preview(markup, text):
        return Page.get_markup_instance(markup).get_document_body(text)

    @property
    def markup_(self):
        if not hasattr(self, '__markup_instance'):
            self.__markup_instance = Page.get_markup_instance(self.markup)
        return self.__markup_instance

    def _get_part(self, part):
        try:
            return getattr(self.markup_, part)(self.raw)
        except SystemMessage:
            return ''

    @property
    def body(self):
        return self.get_cached_content()

    @property
    def stylesheet(self):
        return self._get_part('get_stylesheet')

    @property
    def javascript(self):
        return self._get_part('get_javascript')

    def get_cache_key(self):
        return "waliki:content:%s" % self.slug

    def get_cached_content(self):
        """Returns cached """
        cache_key = self.get_cache_key()
        cached_content = cache.get(cache_key)

        if cached_content is None:
            cached_content = self._get_part('get_document_body')
            cache.set(cache_key, cached_content, WALIKI_CACHE_TIMEOUT)
        return cached_content


class ACLRule(models.Model):
    TO_ANY = 'any'
    TO_LOGGED = 'logged'
    TO_STAFF = 'staff'
    TO_SUPERUSERS = 'superusers'
    TO_EXPLICIT_LIST = 'explicit'
    APPLY_TO_CHOICES = (
        (TO_ANY, _('Any user')),
        (TO_LOGGED, _('Any authenticated user')),
        (TO_STAFF, _('Any staff member')),
        (TO_SUPERUSERS, _('Any superuser')),
        (TO_EXPLICIT_LIST, _('Any user/group explicitly defined')),
    )

    name = models.CharField(verbose_name=_('Name'), max_length=200, unique=True)
    slug = models.CharField(max_length=200)
    as_namespace = models.BooleanField(verbose_name=_('As namespace'),
                                       default=False)
    permissions = models.ManyToManyField(Permission, limit_choices_to={'content_type__app_label': 'waliki'})
    apply_to = models.CharField(max_length=25, choices=APPLY_TO_CHOICES, default=TO_EXPLICIT_LIST)
    users = models.ManyToManyField(settings.AUTH_USER_MODEL, blank=True)
    groups = models.ManyToManyField(Group, blank=True)

    def __unicode__(self):
        return _('Rule: %(name)s for /%(slug)s') % {'name': self.name, 'slug': self.slug}

    def __str__(self):
        return self.__unicode__()

    class Meta:
        verbose_name = _('ACL rule')
        verbose_name_plural = _('ACL rules')

    @classmethod
    def get_rules_for(cls, perms, slug):
        """return rules with ``perms`` for the given ``slug``

        ``perms`` could be the permission name or an iterable
         of permissions names
        """
        def parents_for(slug):
            parent_slugs = []
            for f in slug.rsplit('/'):
                try:
                    parent = parent_slugs[-1] + '/' + f
                except IndexError:
                    parent = f
                parent_slugs.append(parent)
            return parent_slugs

        if isinstance(perms, string_types):
            perms = (perms,)

        rules = cls.objects.all()
        parent_slugs = parents_for(slug)
        for perm in perms:
            lookup = Q(permissions__codename=perm, as_namespace=False, slug=slug)
            for parent in parent_slugs:
                lookup |= Q(permissions__codename=perm, as_namespace=True, slug=parent)
            rules = rules.filter(lookup)
        return rules.distinct()

    @classmethod
    def get_users_for(cls, perms, slug):
        """
        return users with ``perms`` for the given ``slug``.

        ``perms`` could be the permission name or an iterable
         of permissions names
        """

        rules = cls.get_rules_for(perms, slug)
        if rules.filter(apply_to=ACLRule.TO_ANY).exists():
            return {AnonymousUser()} | set(get_user_model().objects.all())
        elif rules.filter(apply_to=ACLRule.TO_LOGGED).exists():
            return get_user_model().objects.all()

        allowed = []
        for rule in rules:
            if rule.apply_to == ACLRule.TO_STAFF:
                allowed += get_user_model().objects.filter(is_staff=True).values_list('id', flat=True)
            elif rule.apply_to == ACLRule.TO_SUPERUSERS:
                allowed += get_user_model().objects.filter(is_superuser=True).values_list('id', flat=True)
            else:
                allowed += get_user_model().objects.filter(Q(aclrule=rule) |
                                                           Q(groups__aclrule=rule)).values_list('id',
                                                                                                flat=True)
        return get_user_model().objects.filter(id__in=allowed).distinct()


class Redirect(models.Model):
    old_slug = models.CharField(max_length=200, unique=True)
    new_slug = models.CharField(max_length=200)
    status_code = models.IntegerField(default=302, choices=((302, '302 Found'), (301, '301 Moved Permanently')))

    def get_absolute_url(self):
        return reverse('waliki_detail', args=(self.new_slug,))


######################################################
# SIGNAL HANDLERS
######################################################

@receiver(post_save, sender=Page)
def on_page_save_clear_cache(instance, **kwargs):
    cache.delete(instance.get_cache_key())


@receiver(pre_delete, sender=Page)
def delete_page(sender, instance, **kwargs):
    if os.path.exists(instance.abspath):
        os.remove(instance.abspath)



# -*- coding: utf-8 -*-
import json
import imghdr
from django.shortcuts import render, get_object_or_404
from django.contrib import messages
from django.utils.six import text_type
from django.http import HttpResponse
from sendfile import sendfile
from waliki.models import Page
from waliki.acl import permission_required
from .models import Attachment


@permission_required('change_page')
def attachments(request, slug):
	last_attached = None
	page = get_object_or_404(Page, slug=slug)
	if request.method == 'POST' and 'attach' in request.FILES:
		last_attached = request.FILES['attach']
		Attachment.objects.create(file=last_attached, page=page)
		messages.success(request, '"%s" was attached succesfully to /%s' % (last_attached.name, page.slug))
	return render(request, 'waliki/attachments.html', {'page': page})


@permission_required('delete_page')
def delete_attachment(request, slug, attachment_id):
	attachment = get_object_or_404(Attachment, id=attachment_id, page__slug=slug)
	name = text_type(attachment)
	if request.is_ajax() and request.method in ('POST', 'DELETE'):
		attachment.delete()
		return HttpResponse(json.dumps({'removed': name}), content_type="application/json")
	return HttpResponse(json.dumps({'removed': None}), content_type="application/json")


@permission_required('view_page', raise_exception=True)
def get_file(request, slug, attachment_id, filename):
	attachment = get_object_or_404(Attachment, id=attachment_id, page__slug=slug)
	as_attachment = ((not imghdr.what(attachment.file.path) and 'embed' not in request.GET)
					  or 'as_attachment' in request.GET)
	# ref https://github.com/johnsensible/django-sendfile
	return sendfile(request, attachment.file.path,
					attachment=as_attachment, attachment_filename=text_type(attachment))


import json
from django.http import HttpResponse, Http404, HttpResponseRedirect, HttpResponsePermanentRedirect
from django.shortcuts import render, redirect, get_object_or_404
from django.template import RequestContext
from django.template.loader import render_to_string
from django.utils.translation import ugettext_lazy as _
from django.contrib import messages
from django.core.urlresolvers import reverse
from .models import Page, Redirect
from .forms import PageForm, MovePageForm, DeleteForm, NewPageForm
from .signals import page_saved, page_preedit, page_moved
from ._markups import get_all_markups
from .acl import permission_required
from . import settings


def home(request):
    return detail(request, slug=settings.WALIKI_INDEX_SLUG)


@permission_required('view_page')
def detail(request, slug, raw=False):

    slug = slug.strip('/')

    # handle redirects first
    try:
        redirect = Redirect.objects.get(old_slug=slug)      # noqa
        if redirect.status_code == 302:
            return HttpResponseRedirect(redirect.get_absolute_url())
        return HttpResponsePermanentRedirect(redirect.get_absolute_url())
    except Redirect.DoesNotExist:
        pass

    try:
        page = Page.objects.get(slug=slug)
    except Page.DoesNotExist:
        page = None
    if raw and page:
        return HttpResponse(page.raw, content_type='text/plain; charset=utf-8')
    elif raw:
        raise Http404

    return render(request, 'waliki/detail.html', {'page': page, 'slug': slug})


@permission_required('change_page')
def move(request, slug):
    page = get_object_or_404(Page, slug=slug)
    data = request.POST if request.method == 'POST' else None
    form = MovePageForm(data, instance=page)
    if request.method == 'POST' and form.is_valid():
        new_slug = form.cleaned_data['slug']

        # remove redirections (now there is an actual page)
        Redirect.objects.filter(old_slug=new_slug).delete()
        # squash redirections to new destiny
        Redirect.objects.filter(new_slug=slug).update(new_slug=new_slug)
        # create the new redirection
        Redirect.objects.create(old_slug=slug, new_slug=new_slug)

        if not form.cleaned_data['just_redirect']:
            old_path = page.path
            page.move(new_slug + page.markup_.file_extensions[0])
            page.slug = new_slug
            page.save()
            msg = _("Page moved from %(old_slug)s") % {'old_slug': slug}
            page_moved.send(sender=move,
                            page=page,
                            old_path=old_path,
                            author=request.user,
                            message=msg,
                            )
            url = page.get_absolute_url()
        else:
            msg = _("A redirection from %(old_slug)s to this page was added") % {'old_slug': slug}

        url = reverse('waliki_detail', args=[new_slug])
        messages.success(request, msg)

        if request.is_ajax():
            return HttpResponse(json.dumps({'redirect': url}), content_type="application/json")
        return redirect(url)

    if request.is_ajax():
        data = render_to_string('waliki/generic_form.html', {'page': page, 'form': form},
                                context_instance=RequestContext(request))
        return HttpResponse(json.dumps({'data': data}), content_type="application/json")
    return render(request, 'waliki/generic_form.html', {'page': page, 'form': form})


@permission_required('change_page')
def edit(request, slug):
    slug = slug.strip('/')
    just_created = False
    try:
        page = Page.objects.get(slug=slug)
    except Page.DoesNotExist:
        if request.method == 'POST':
            page = Page.objects.create(slug=slug)
            page.raw = ""
            page_saved.send(sender=edit,
                            page=page,
                            author=request.user,
                            message=_("Page created"),
                            form_extra_data={})
            just_created = True
        else:
            return redirect('waliki_detail', slug)

    original_markup = page.markup
    data = request.POST if request.method == 'POST' and not just_created else None
    form_extra_data = {}
    receivers_responses = page_preedit.send(sender=edit, page=page)
    for r in receivers_responses:
        if isinstance(r[1], dict) and 'form_extra_data' in r[1]:
            form_extra_data.update(r[1]['form_extra_data'])

    form = PageForm(
        data, instance=page, initial={'extra_data': json.dumps(form_extra_data)})
    if form.is_valid():
        page = form.save(commit=False)
        if page.markup != original_markup:
            old_path = page.path
            page.update_extension()
            msg = _("The markup was changed from {original} to {new}").format(original=original_markup,
                                                                              new=page.markup)
            page_moved.send(sender=edit,
                            page=page,
                            old_path=old_path,
                            author=request.user,
                            message=msg,
                            commit=False
                            )
            was_moved = True
            messages.warning(request, msg)
        else:
            was_moved = False
        page.raw = form.cleaned_data['raw']
        page.save()
        try:
            receivers_responses = page_saved.send(sender=edit,
                                                  page=page,
                                                  author=request.user,
                                                  message=form.cleaned_data[
                                                      "message"],
                                                  form_extra_data=json.loads(form.cleaned_data["extra_data"] or "{}"),
                                                  was_moved=was_moved)
        except Page.EditionConflict as e:
            messages.warning(request, e)
            return redirect('waliki_edit', slug=page.slug)

        for r in receivers_responses:
            if isinstance(r[1], dict) and 'messages' in r[1]:
                for key, value in r[1]['messages'].items():
                    getattr(messages, key)(request, value)

        if 'next' in request.GET:
            return redirect(request.GET['next'])
        return redirect('waliki_detail', slug=page.slug)
    cm_modes = [(m.name, m.codemirror_mode_name) for m in get_all_markups()]

    cm_settings = settings.WALIKI_CODEMIRROR_SETTINGS
    cm_settings.update({'mode': dict(cm_modes)[page.markup]})
    return render(request, 'waliki/edit.html', {'page': page,
                                                'form': form,
                                                'slug': slug,
                                                'cm_modes': cm_modes,
                                                'cm_settings': json.dumps(cm_settings)})


def preview(request):
    data = {}
    if request.is_ajax() and request.method == "POST":
        data['html'] = Page.preview(
            request.POST['markup'], request.POST['text'])
        return HttpResponse(json.dumps(data), content_type="application/json")


@permission_required('delete_page')
def delete(request, slug):
    page = get_object_or_404(Page, slug=slug)
    data = request.POST if request.method == 'POST' else None
    form = DeleteForm(data)
    if form.is_valid():
        if form.cleaned_data['what'] == 'this':
            msg = _("The page %(slug)s was deleted") % {'slug': slug}
            page.delete()
        else:
            Page.objects.filter(slug__startswith=slug).delete()
            msg = _("The page %(slug)s and all its namespace was deleted") % {
                'slug': slug}

        messages.warning(request, msg)
        if request.is_ajax():
            return HttpResponse(json.dumps({'redirect': reverse('waliki_home')}), content_type="application/json")
        return redirect('waliki_home')

    if request.is_ajax():
        data = render_to_string('waliki/delete.html', {'page': page, 'form': form},
                                context_instance=RequestContext(request))
        return HttpResponse(json.dumps({'data': data}), content_type="application/json")
    return render(request, 'waliki/delete.html', {'page': page, 'form': form})


def new(request):
    data = request.POST if request.method == 'POST' else None
    form = NewPageForm(data, user=request.user)
    if request.method == 'POST' and form.is_valid():
        page = form.save()
        page.raw = ""
        page_saved.send(sender=new,
                        page=page,
                        author=request.user,
                        message=_("Page created"),
                        form_extra_data={})
        if request.is_ajax():
            return HttpResponse(json.dumps({'redirect': page.get_edit_url()}), content_type="application/json")
        return redirect(page.get_edit_url())

    if request.is_ajax():
        data = render_to_string('waliki/generic_form.html', {'form': form},
                                context_instance=RequestContext(request))
        return HttpResponse(json.dumps({'data': data}), content_type="application/json")
    return render(request, 'waliki/generic_form.html', {'form': form})


def get_slug(request):
    slug = settings.get_slug(request.GET.get('title', ''))
    return HttpResponse(json.dumps({'slug': slug}), content_type="application/json")



from os import path
import shutil
import tempfile
from sh import hovercraft
from django.shortcuts import get_object_or_404
from django.http import HttpResponse
from waliki.models import Page
from waliki.acl import permission_required


@permission_required('view_page')
def slides(request, slug):
    page = get_object_or_404(Page, slug=slug)
    outpath = tempfile.mkdtemp()
    try:
        infile = page.abspath
        template = path.join(path.dirname(path.realpath(__file__)), 'template')
        hovercraft('-t', template, infile, outpath)
        with open(path.join(outpath, 'index.html')) as f:
            content = f.read()
    finally:
        shutil.rmtree(outpath)
    return HttpResponse(content)



from django.contrib import admin
from .models import Page, ACLRule, Redirect


class PageAdmin(admin.ModelAdmin):
    list_display = ('slug', 'title', 'markup')
    list_filter = ('markup',)
    search_fields = ('slug', 'title')


class ACLRuleAdmin(admin.ModelAdmin):
    list_display = ('name', 'slug', 'as_namespace', 'apply_to')
    list_filter = ('apply_to', 'permissions')
    search_fields = ('slug', 'name', 'users__username', 'groups__name')


class RedirectAdmin(admin.ModelAdmin):
    list_display = ('old_slug', 'new_slug', 'status_code')
    search_fields = ('old_slug', 'new_slug')


# Register your models here.
admin.site.register(Page, PageAdmin)
admin.site.register(ACLRule, ACLRuleAdmin)
admin.site.register(Redirect, RedirectAdmin)



from django.utils.translation import ugettext_lazy as _
from waliki.plugins import BasePlugin, register


class SlidesPlugin(BasePlugin):
    slug = 'slides'
    urls_page = ['waliki.slides.urls']
    extra_page_actions = {'reStructuredText': [('waliki_slides', _('View as slides'))]}

register(SlidesPlugin)




from django.conf.urls import patterns, url
from waliki.settings import WALIKI_SLUG_PATTERN


urlpatterns = patterns('waliki.slides.views',
    url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/slides$', 'slides', name='waliki_slides'),
)



import json
from django.shortcuts import render, get_object_or_404, redirect
from django.utils.translation import ugettext_lazy as _
from django.http import Http404
from django.core.management import call_command
from django.http import HttpResponse
from django.core.urlresolvers import reverse
from django.utils.encoding import smart_text
from django.utils.six import StringIO, text_type
from django.views.decorators.csrf import csrf_exempt
from waliki.models import Page
from waliki.forms import PageForm
from waliki.acl import permission_required
from waliki import settings
from . import Git


@permission_required('view_page')
def history(request, slug, pag=1):
    page = get_object_or_404(Page, slug=slug)
    # The argument passed for pag might be a string, but we want to
    # do calculations on it. So we must cast just to be sure.
    pag = int(pag or 1)
    skip = (pag - 1) * settings.WALIKI_PAGINATE_BY
    max_count = settings.WALIKI_PAGINATE_BY
    if request.method == 'POST':
        new, old = request.POST.getlist('commit')
        return redirect('waliki_diff', slug, old, new)
    history = Git().history(page)
    max_changes = max([(v['insertion'] + v['deletion']) for v in history])
    return render(request, 'waliki/history.html', {'page': page,
                                                   'slug': slug,
                                                   'history': history[skip:(skip+max_count)],
                                                   'max_changes': max_changes,
                                                   'prev': pag - 1 if pag > 1 else None,
                                                   'next': pag + 1 if skip + max_count < len(history) else None})


@permission_required('view_page')
def version(request, slug, version, raw=False):
    page = get_object_or_404(Page, slug=slug)
    content = Git().version(page, version)
    if not content:
        raise Http404
    form = PageForm(instance=page, initial={'message': _('Restored version @%s') % version, 'raw': content},
                    is_hidden=True)

    if raw:
        return HttpResponse(content, content_type='text/plain; charset=utf-8')

    content = Page.preview(page.markup, content)
    return render(request, 'waliki/version.html', {'page': page,
                                                   'content': content,
                                                   'slug': slug,
                                                   'version': version,
                                                   'form': form})


@permission_required('view_page')
def diff(request, slug, old, new, raw=False):
    page = get_object_or_404(Page, slug=slug)
    if raw:
        content = Git().diff(page, new, old)
        return HttpResponse(content, content_type='text/plain; charset=utf-8')
    space = smart_text(b'\xc2\xa0', encoding='utf-8')  # non-breaking space character
    tab = space * 4
    old_content = Git().version(page, old).replace('\t', tab).replace(' ', space)
    new_content = Git().version(page, new).replace('\t', tab).replace(' ', space)
    return render(request, 'waliki/diff.html', {'page': page,
                                                'old_content': old_content,
                                                'new_content': new_content,
                                                'slug': slug,
                                                'old_commit': old,
                                                'new_commit': new})


def whatchanged(request, pag=1):
    changes = []
    # The argument passed for pag might be a string, but we want to
    # do calculations on it. So we must cast just to be sure.
    pag = int(pag or 1)
    skip = (pag - 1) * settings.WALIKI_PAGINATE_BY
    max_count = settings.WALIKI_PAGINATE_BY
    # Git().total_commits() returns a unicode string
    # but we want to do calculations on the number it represents,
    # therefore, we cast
    total = int(Git().total_commits())
    for version in Git().whatchanged(skip, max_count):
        for path in version[-1]:
            try:
                page = Page.objects.get(path=path)
            except Page.DoesNotExist:
                continue
            changes.append({'page': page, 'author': version[0],
                            'version': version[2], 'message': version[3],
                            'date': version[4]})

    return render(request, 'waliki/whatchanged.html', {'changes': changes,
                                                       'prev': pag - 1 if pag > 1 else None,
                                                       'next': pag + 1 if skip + max_count < total else None})


@csrf_exempt
def webhook_pull(request, remote='origin'):
    if request.method == 'POST':
        try:
            log = Git().pull(remote)
            s = StringIO()
            call_command('sync_waliki', stdout=s)
            s.seek(0)
            r = {'pull': log, 'sync': s.read()}
            status_code = 200
        except Exception as e:
            r = {'error': text_type(e)}
            status_code = 500
        return HttpResponse(json.dumps(r), status=status_code,
                            content_type="application/json")

    return HttpResponse("POST to %s" % reverse("waliki_webhook_pull", args=(remote,)))



from django.utils.translation import ugettext_lazy as _
from waliki.plugins import BasePlugin, register


class AttachmentsPlugin(BasePlugin):
    slug = 'attachments'
    urls_page = ['waliki.attachments.urls']
    extra_edit_actions = {'all': [('waliki_attachments', _('Attachments'))]}
    includes_edit = {'content': 'waliki/attachments_modal.html',
    				 'extra_script': 'waliki/attachments_js.html'}

register(AttachmentsPlugin)




from docutils import nodes
from docutils.transforms import Transform
from docutils.readers.standalone import Reader

from waliki.plugins import str2object
from waliki.settings import WALIKI_RST_TRANSFORMS


class Emojis(Transform):
    """
    Replace some substitutions if they aren't defined in the document.
    """
    # run before the default Substitutions
    default_priority = 210

    # list from http://www.tortue.me/
    emojis = set([
                'bowtie',      # NOQA
                'smile',
                'laughing',
                'blush',
                'smiley',
                'relaxed',
                'smirk',
                'heart_eyes',
                'kissing_heart',
                'kissing_closed_eyes',
                'flushed',
                'relieved',
                'satisfied',
                'grin',
                'wink',
                'stuck_out_tongue_winking_eye',
                'stuck_out_tongue_closed_eyes',
                'grinning',
                'kissing',
                'kissing_smiling_eyes',
                'stuck_out_tongue',
                'sleeping',
                'worried',
                'frowning',
                'anguished',
                'open_mouth',
                'grimacing',
                'confused',
                'hushed',
                'expressionless',
                'unamused',
                'sweat_smile',
                'sweat',
                'weary',
                'pensive',
                'disappointed',
                'confounded',
                'fearful',
                'cold_sweat',
                'persevere',
                'cry',
                'sob',
                'joy',
                'astonished',
                'scream',
                'neckbeard',
                'tired_face',
                'angry',
                'rage',
                'triumph',
                'sleepy',
                'yum',
                'mask',
                'sunglasses',
                'dizzy_face',
                'imp',
                'smiling_imp',
                'neutral_face',
                'no_mouth',
                'innocent',
                'alien',
                'yellow_heart',
                'blue_heart',
                'purple_heart',
                'heart',
                'green_heart',
                'broken_heart',
                'heartbeat',
                'heartpulse',
                'two_hearts',
                'revolving_hearts',
                'cupid',
                'sparkling_heart',
                'sparkles',
                'star',
                'star2',
                'dizzy',
                'boom',
                'collision',
                'anger',
                'exclamation',
                'question',
                'grey_exclamation',
                'grey_question',
                'zzz',
                'dash',
                'sweat_drops',
                'notes',
                'musical_note',
                'fire',
                'hankey',
                'poop',
                'shit',
                '+1',
                'thumbsup',
                '-1',
                'thumbsdown',
                'ok_hand',
                'punch',
                'facepunch',
                'fist',
                'v',
                'wave',
                'hand',
                'open_hands',
                'point_up',
                'point_down',
                'point_left',
                'point_right',
                'raised_hands',
                'pray',
                'point_up_2',
                'clap',
                'muscle',
                'metal',
                'walking',
                'runner',
                'running',
                'couple',
                'family',
                'two_men_holding_hands',
                'two_women_holding_hands',
                'dancer',
                'dancers',
                'ok_woman',
                'no_good',
                'information_desk_person',
                'raised_hand',
                'bride_with_veil',
                'person_with_pouting_face',
                'person_frowning',
                'bow',
                'couplekiss',
                'couple_with_heart',
                'massage',
                'haircut',
                'nail_care',
                'boy',
                'girl',
                'woman',
                'man',
                'baby',
                'older_woman',
                'older_man',
                'person_with_blond_hair',
                'man_with_gua_pi_mao',
                'man_with_turban',
                'construction_worker',
                'cop',
                'angel',
                'princess',
                'smiley_cat',
                'smile_cat',
                'heart_eyes_cat',
                'kissing_cat',
                'smirk_cat',
                'scream_cat',
                'crying_cat_face',
                'joy_cat',
                'pouting_cat',
                'japanese_ogre',
                'japanese_goblin',
                'see_no_evil',
                'hear_no_evil',
                'speak_no_evil',
                'guardsman',
                'skull',
                'feet',
                'lips',
                'kiss',
                'droplet',
                'ear',
                'eyes',
                'nose',
                'tongue',
                'love_letter',
                'bust_in_silhouette',
                'busts_in_silhouette',
                'speech_balloon',
                'thought_balloon',
                'feelsgood',
                'finnadie',
                'goberserk',
                'godmode',
                'hurtrealbad',
                'rage1',
                'rage2',
                'rage3',
                'rage4',
                'suspect',
                'trollface',
                'sunny',
                'umbrella',
                'cloud',
                'snowflake',
                'snowman',
                'zap',
                'cyclone',
                'foggy',
                'ocean',
                'cat',
                'dog',
                'mouse',
                'hamster',
                'rabbit',
                'wolf',
                'frog',
                'tiger',
                'koala',
                'bear',
                'pig',
                'pig_nose',
                'cow',
                'boar',
                'monkey_face',
                'monkey',
                'horse',
                'racehorse',
                'camel',
                'sheep',
                'elephant',
                'panda_face',
                'snake',
                'bird',
                'baby_chick',
                'hatched_chick',
                'hatching_chick',
                'chicken',
                'penguin',
                'turtle',
                'bug',
                'honeybee',
                'ant',
                'beetle',
                'snail',
                'octopus',
                'tropical_fish',
                'fish',
                'whale',
                'whale2',
                'dolphin',
                'cow2',
                'ram',
                'rat',
                'water_buffalo',
                'tiger2',
                'rabbit2',
                'dragon',
                'goat',
                'rooster',
                'dog2',
                'pig2',
                'mouse2',
                'ox',
                'dragon_face',
                'blowfish',
                'crocodile',
                'dromedary_camel',
                'leopard',
                'cat2',
                'poodle',
                'paw_prints',
                'bouquet',
                'cherry_blossom',
                'tulip',
                'four_leaf_clover',
                'rose',
                'sunflower',
                'hibiscus',
                'maple_leaf',
                'leaves',
                'fallen_leaf',
                'herb',
                'mushroom',
                'cactus',
                'palm_tree',
                'evergreen_tree',
                'deciduous_tree',
                'chestnut',
                'seedling',
                'blossom',
                'ear_of_rice',
                'shell',
                'globe_with_meridians',
                'sun_with_face',
                'full_moon_with_face',
                'new_moon_with_face',
                'new_moon',
                'waxing_crescent_moon',
                'first_quarter_moon',
                'waxing_gibbous_moon',
                'full_moon',
                'waning_gibbous_moon',
                'last_quarter_moon',
                'waning_crescent_moon',
                'last_quarter_moon_with_face',
                'first_quarter_moon_with_face',
                'moon',
                'earth_africa',
                'earth_americas',
                'earth_asia',
                'volcano',
                'milky_way',
                'partly_sunny',
                'octocat',
                'squirrel',
                'bamboo',
                'gift_heart',
                'dolls',
                'school_satchel',
                'mortar_board',
                'flags',
                'fireworks',
                'sparkler',
                'wind_chime',
                'rice_scene',
                'jack_o_lantern',
                'ghost',
                'santa',
                'christmas_tree',
                'gift',
                'bell',
                'no_bell',
                'tanabata_tree',
                'tada',
                'confetti_ball',
                'balloon',
                'crystal_ball',
                'cd',
                'dvd',
                'floppy_disk',
                'camera',
                'video_camera',
                'movie_camera',
                'computer',
                'tv',
                'iphone',
                'phone',
                'telephone',
                'telephone_receiver',
                'pager',
                'fax',
                'minidisc',
                'vhs',
                'sound',
                'speaker',
                'mute',
                'loudspeaker',
                'mega',
                'hourglass',
                'hourglass_flowing_sand',
                'alarm_clock',
                'watch',
                'radio',
                'satellite',
                'loop',
                'mag',
                'mag_right',
                'unlock',
                'lock',
                'lock_with_ink_pen',
                'closed_lock_with_key',
                'key',
                'bulb',
                'flashlight',
                'high_brightness',
                'low_brightness',
                'electric_plug',
                'battery',
                'calling',
                'email',
                'mailbox',
                'postbox',
                'bath',
                'bathtub',
                'shower',
                'toilet',
                'wrench',
                'nut_and_bolt',
                'hammer',
                'seat',
                'moneybag',
                'yen',
                'dollar',
                'pound',
                'euro',
                'credit_card',
                'money_with_wings',
                'e-mail',
                'inbox_tray',
                'outbox_tray',
                'envelope',
                'incoming_envelope',
                'postal_horn',
                'mailbox_closed',
                'mailbox_with_mail',
                'mailbox_with_no_mail',
                'door',
                'smoking',
                'bomb',
                'gun',
                'hocho',
                'pill',
                'syringe',
                'page_facing_up',
                'page_with_curl',
                'bookmark_tabs',
                'bar_chart',
                'chart_with_upwards_trend',
                'chart_with_downwards_trend',
                'scroll',
                'clipboard',
                'calendar',
                'date',
                'card_index',
                'file_folder',
                'open_file_folder',
                'scissors',
                'pushpin',
                'paperclip',
                'black_nib',
                'pencil2',
                'straight_ruler',
                'triangular_ruler',
                'closed_book',
                'green_book',
                'blue_book',
                'orange_book',
                'notebook',
                'notebook_with_decorative_cover',
                'ledger',
                'books',
                'bookmark',
                'name_badge',
                'microscope',
                'telescope',
                'newspaper',
                'football',
                'basketball',
                'soccer',
                'baseball',
                'tennis',
                '8ball',
                'rugby_football',
                'bowling',
                'golf',
                'mountain_bicyclist',
                'bicyclist',
                'horse_racing',
                'snowboarder',
                'swimmer',
                'surfer',
                'ski',
                'spades',
                'hearts',
                'clubs',
                'diamonds',
                'gem',
                'ring',
                'trophy',
                'musical_score',
                'musical_keyboard',
                'violin',
                'space_invader',
                'video_game',
                'black_joker',
                'flower_playing_cards',
                'game_die',
                'dart',
                'mahjong',
                'clapper',
                'memo',
                'pencil',
                'book',
                'art',
                'microphone',
                'headphones',
                'trumpet',
                'saxophone',
                'guitar',
                'shoe',
                'sandal',
                'high_heel',
                'lipstick',
                'boot',
                'shirt',
                'tshirt',
                'necktie',
                'womans_clothes',
                'dress',
                'running_shirt_with_sash',
                'jeans',
                'kimono',
                'bikini',
                'ribbon',
                'tophat',
                'crown',
                'womans_hat',
                'mans_shoe',
                'closed_umbrella',
                'briefcase',
                'handbag',
                'pouch',
                'purse',
                'eyeglasses',
                'fishing_pole_and_fish',
                'coffee',
                'tea',
                'sake',
                'baby_bottle',
                'beer',
                'beers',
                'cocktail',
                'tropical_drink',
                'wine_glass',
                'fork_and_knife',
                'pizza',
                'hamburger',
                'fries',
                'poultry_leg',
                'meat_on_bone',
                'spaghetti',
                'curry',
                'fried_shrimp',
                'bento',
                'sushi',
                'fish_cake',
                'rice_ball',
                'rice_cracker',
                'rice',
                'ramen',
                'stew',
                'oden',
                'dango',
                'egg',
                'bread',
                'doughnut',
                'custard',
                'icecream',
                'ice_cream',
                'shaved_ice',
                'birthday',
                'cake',
                'cookie',
                'chocolate_bar',
                'candy',
                'lollipop',
                'honey_pot',
                'apple',
                'green_apple',
                'tangerine',
                'lemon',
                'cherries',
                'grapes',
                'watermelon',
                'strawberry',
                'peach',
                'melon',
                'banana',
                'pear',
                'pineapple',
                'sweet_potato',
                'eggplant',
                'tomato',
                'corn',
                '109',
                'house',
                'house_with_garden',
                'school',
                'office',
                'post_office',
                'hospital',
                'bank',
                'convenience_store',
                'love_hotel',
                'hotel',
                'wedding',
                'church',
                'department_store',
                'european_post_office',
                'city_sunrise',
                'city_sunset',
                'japanese_castle',
                'european_castle',
                'tent',
                'factory',
                'tokyo_tower',
                'japan',
                'mount_fuji',
                'sunrise_over_mountains',
                'sunrise',
                'stars',
                'statue_of_liberty',
                'bridge_at_night',
                'carousel_horse',
                'rainbow',
                'ferris_wheel',
                'fountain',
                'roller_coaster',
                'ship',
                'speedboat',
                'boat',
                'sailboat',
                'rowboat',
                'anchor',
                'rocket',
                'airplane',
                'helicopter',
                'steam_locomotive',
                'tram',
                'mountain_railway',
                'bike',
                'aerial_tramway',
                'suspension_railway',
                'mountain_cableway',
                'tractor',
                'blue_car',
                'oncoming_automobile',
                'car',
                'red_car',
                'taxi',
                'oncoming_taxi',
                'articulated_lorry',
                'bus',
                'oncoming_bus',
                'rotating_light',
                'police_car',
                'oncoming_police_car',
                'fire_engine',
                'ambulance',
                'minibus',
                'truck',
                'train',
                'station',
                'train2',
                'bullettrain_front',
                'bullettrain_side',
                'light_rail',
                'monorail',
                'railway_car',
                'trolleybus',
                'ticket',
                'fuelpump',
                'vertical_traffic_light',
                'traffic_light',
                'warning',
                'construction',
                'beginner',
                'atm',
                'slot_machine',
                'busstop',
                'barber',
                'hotsprings',
                'checkered_flag',
                'crossed_flags',
                'izakaya_lantern',
                'moyai',
                'circus_tent',
                'performing_arts',
                'round_pushpin',
                'triangular_flag_on_post',
                'jp',
                'kr',
                'cn',
                'us',
                'fr',
                'es',
                'it',
                'ru',
                'gb',
                'uk',
                'de',
                'one',
                'two',
                'three',
                'four',
                'five',
                'six',
                'seven',
                'eight',
                'nine',
                'keycap_ten',
                '1234',
                'zero',
                'hash',
                'symbols',
                'arrow_backward',
                'arrow_down',
                'arrow_forward',
                'arrow_left',
                'capital_abcd',
                'abcd',
                'abc',
                'arrow_lower_left',
                'arrow_lower_right',
                'arrow_right',
                'arrow_up',
                'arrow_upper_left',
                'arrow_upper_right',
                'arrow_double_down',
                'arrow_double_up',
                'arrow_down_small',
                'arrow_heading_down',
                'arrow_heading_up',
                'leftwards_arrow_with_hook',
                'arrow_right_hook',
                'left_right_arrow',
                'arrow_up_down',
                'arrow_up_small',
                'arrows_clockwise',
                'arrows_counterclockwise',
                'rewind',
                'fast_forward',
                'information_source',
                'ok',
                'twisted_rightwards_arrows',
                'repeat',
                'repeat_one',
                'new',
                'top',
                'up',
                'cool',
                'free',
                'ng',
                'cinema',
                'koko',
                'signal_strength',
                'u5272',
                'u5408',
                'u55b6',
                'u6307',
                'u6708',
                'u6709',
                'u6e80',
                'u7121',
                'u7533',
                'u7a7a',
                'u7981',
                'sa',
                'restroom',
                'mens',
                'womens',
                'baby_symbol',
                'no_smoking',
                'parking',
                'wheelchair',
                'metro',
                'baggage_claim',
                'accept',
                'wc',
                'potable_water',
                'put_litter_in_its_place',
                'secret',
                'congratulations',
                'm',
                'passport_control',
                'left_luggage',
                'customs',
                'ideograph_advantage',
                'cl',
                'sos',
                'id',
                'no_entry_sign',
                'underage',
                'no_mobile_phones',
                'do_not_litter',
                'non-potable_water',
                'no_bicycles',
                'no_pedestrians',
                'children_crossing',
                'no_entry',
                'eight_spoked_asterisk',
                'eight_pointed_black_star',
                'heart_decoration',
                'vs',
                'vibration_mode',
                'mobile_phone_off',
                'chart',
                'currency_exchange',
                'aries',
                'taurus',
                'gemini',
                'cancer',
                'leo',
                'virgo',
                'libra',
                'scorpius',
                'sagittarius',
                'capricorn',
                'aquarius',
                'pisces',
                'ophiuchus',
                'six_pointed_star',
                'negative_squared_cross_mark',
                'a',
                'b',
                'ab',
                'o2',
                'diamond_shape_with_a_dot_inside',
                'recycle',
                'end',
                'true',
                'soon',
                'clock1',
                'clock130',
                'clock10',
                'clock1030',
                'clock11',
                'clock1130',
                'clock12',
                'clock1230',
                'clock2',
                'clock230',
                'clock3',
                'clock330',
                'clock4',
                'clock430',
                'clock5',
                'clock530',
                'clock6',
                'clock630',
                'clock7',
                'clock730',
                'clock8',
                'clock830',
                'clock9',
                'clock930',
                'heavy_dollar_sign',
                'copyright',
                'registered',
                'tm',
                'x',
                'heavy_exclamation_mark',
                'bangbang',
                'interrobang',
                'o',
                'heavy_multiplication_x',
                'heavy_plus_sign',
                'heavy_minus_sign',
                'heavy_division_sign',
                'white_flower',
                '100',
                'heavy_check_mark',
                'ballot_box_with_check',
                'radio_button',
                'link',
                'curly_loop',
                'wavy_dash',
                'part_alternation_mark',
                'trident',
                'black_square',
                'white_square',
                'white_check_mark',
                'black_square_button',
                'white_square_button',
                'black_circle',
                'white_circle',
                'red_circle',
                'large_blue_circle',
                'large_blue_diamond',
                'large_orange_diamond',
                'small_blue_diamond',
                'small_orange_diamond',
                'small_red_triangle',
                'small_red_triangle_down',
                'shipit'])

    def apply(self, **kwargs):
        # only handle those not otherwise defined in the document
        to_handle = self.emojis - set(self.document.substitution_defs)
        for ref in self.document.traverse(nodes.substitution_reference):
            refname = ref['refname']
            if refname in to_handle:
                node = nodes.image(
                    uri='http://www.tortue.me/emoji/{0}.png'.format(refname),
                    alt=refname,
                    classes=['emoji'],
                    height="24px",
                    width="24px")
                ref.replace_self(node)


class WalikiReader(Reader):

    def get_transforms(self):
        transforms = [str2object(transform) for transform in WALIKI_RST_TRANSFORMS]
        return Reader.get_transforms(self) + transforms



# -*- coding: utf-8 -*-

# taken from
# https://github.com/pinax/django-forms-bootstrap

from django import template
from django.conf import settings
from django.template import Context
from django.template.loader import get_template


register = template.Library()


def _preprocess_fields(form):
    for field in form.fields:
        name = form.fields[field].widget.__class__.__name__.lower()
        if not name.startswith("radio") and not name.startswith("checkbox"):
            try:
                form.fields[field].widget.attrs["class"] += " form-control"
            except KeyError:
                form.fields[field].widget.attrs["class"] = " form-control"
    return form


@register.filter
def render_form(form):
    """same than  {{ form|crispy }} if crispy_forms is installed.
    render using a bootstrap3 templating otherwise"""

    if 'crispy_forms' in settings.INSTALLED_APPS:
        from crispy_forms.templatetags.crispy_forms_filters import as_crispy_form
        return as_crispy_form(form)

    template = get_template("bootstrap/form.html")
    form = _preprocess_fields(form)

    c = Context({
        "form": form,
    })
    return template.render(c)


@register.filter
def as_bootstrap_inline(form):
    template = get_template("bootstrap/form.html")
    form = _preprocess_fields(form)

    for field in form.fields:
        name = form.fields[field].widget.__class__.__name__.lower()
        if not name.startswith("radio") and not name.startswith("checkbox"):
            form.fields[field].widget.attrs["placeholder"] = form.fields[field].label

    css_classes = {
        "label": "sr-only",
        "single_container": "",
        "wrap": "",
    }

    c = Context({
        "form": form,
        "css_classes": css_classes,
    })
    return template.render(c)


@register.filter
def as_bootstrap_horizontal(form, label_classes=""):
    template = get_template("bootstrap/form.html")
    form = _preprocess_fields(form)

    if label_classes == "":
        label_classes = "col-md-2"

    css_classes = {
        "label": label_classes,
        "single_container": "",
        "wrap": "",
    }

    for label_class in label_classes.split(" "):
        split_class, column_count = label_class.rsplit("-", 1)
        column_count = int(column_count)

        if column_count < 12:
            offset_class = "{split_class}-offset-{column_count}".format(
                split_class=split_class,
                column_count=column_count,
            )
            wrap_class = "{split_class}-{column_count}".format(
                split_class=split_class,
                column_count=12 - column_count,
            )
            css_classes["single_container"] += offset_class + " " + wrap_class + " "
            css_classes["wrap"] += wrap_class + " "

    c = Context({
        "form": form,
        "css_classes": css_classes,
    })
    return template.render(c)


@register.filter
def css_class(field):
    return field.field.widget.__class__.__name__.lower()



from django.conf.urls import patterns, url
from waliki.settings import WALIKI_SLUG_PATTERN

urlpatterns = patterns('waliki.attachments.views',
    url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/attachments$', 'attachments',
        name='waliki_attachments'),
    url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/attachments/(?P<attachment_id>\d+)/delete$',
        'delete_attachment', name='waliki_delete_attachment'),
    url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/attachment/(?P<attachment_id>\d+)/(?P<filename>.*)$',
        'get_file',   name='waliki_attachment_file')
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
this is the module from
https://bitbucket.org/andre_felipe_dias/rst2html5
"""


from __future__ import unicode_literals

__docformat__ = 'reStructuredText'

import re
import sys
from docutils import nodes, writers, frontend
from docutils.transforms import Transform
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
try:
    # docutils >= 0.10
    from docutils.utils.math import pick_math_environment
except ImportError:
    from docutils.math import pick_math_environment
from genshi.builder import tag
from genshi.output import XHTMLSerializer
from genshi.core import Markup

if sys.version[0] == '3':
    unicode = str


class FooterToBottom(Transform):
    '''
    The doctree must be adjusted before translation begins
    since in-place tree modifications doesn't work.

    The footer must be relocated to the bottom of the
    doctree in order to be translated at the right position.

    .. seealso::

        * http://docutils.sourceforge.net/docs/ref/transforms.html
    '''

    default_priority = 850

    def apply(self):
        for child in self.document.children:
            if isinstance(child, nodes.decoration):
                for footer in child:
                    if isinstance(footer, nodes.footer):
                        child.remove(footer)
                        self.document.append(footer)
                break


def _script_callback_option_parser(option, opt_str, value, parser):
    '''
    Store a script's URL or path along to its attribute 'defer', 'async' or None
    '''
    attr = 'defer' in opt_str and 'defer' or \
           'async' in opt_str and 'async' or \
           None
    parser.values.script = parser.values.script or []
    parser.values.script.append((value, attr))
    return


class HTML5Writer(writers.Writer):

    supported = ('html', 'html5', 'html5css3')

    config_section = 'html5writer'
    config_section_dependencies = ('writers')

    # OptParse see https://docs.python.org/2/library/optparse.html
    settings_spec = (
        'HTML5 Specific Options',
        None, (
        ("Don't indent output", ['--no-indent'],
            {'default': 1, 'action': 'store_false', 'dest': 'indent_output'}),
        ('Specify a stylesheet URL to be included in the output HTML file. '
         '(This option can be used multiple times)',
            ['--stylesheet'],
            {'metavar': '<URL or path>', 'default': None,
             'action': 'append', }),
        ('Specify a script URL to be included in the output HTML file. '
         '(This option can be used multiple times)',
            ['--script'],
            {'metavar': '<URL or path>', 'default': None,
             'dest': 'script',
             'type': 'string',
             'action': 'callback',
             'callback': _script_callback_option_parser, }),
        ('Specify a script URL with a defer attribute '
         'to be included in the output HTML file. '
         '(This option can be used multiple times)',
            ['--script-defer'],
            {'metavar': '<URL or path>',
             'dest': 'script',
             'type': 'string',
             'action': 'callback',
             'callback': _script_callback_option_parser, }),
        ('Specify a script URL with a async attribute '
         'to be included in the output HTML file. '
         '(This option can be used multiple times)',
            ['--script-async'],
            {'metavar': '<URL or path>',
             'dest': 'script',
             'type': 'string',
             'action': 'callback',
             'callback': _script_callback_option_parser, }),
        ('Specify a html tag attribute. '
         '(This option can be used multiple times)',
            ['--html-tag-attr'],
            {'metavar': '<attribute>', 'default': None,
             'dest': 'html_tag_attr',
             'action': 'append', }),
        ('Specify a filename or text to be used as the HTML5 output template. '
         'The template must have the {head} and {body} placeholders. '
         'The "<html{html_attr}>" placeholder is recommended.',
            ['--template'],
            {'metavar': '<filename or text>', 'default': None,
             'dest': 'template',
             'type': 'string',
             'action': 'store', }),
        )
    )

    settings_defaults = {
        'tab_width': 4,
        'syntax_highlight': 'short',
        'field_limit': 0,
        'option_limit': 0,
    }

    def __init__(self):
        writers.Writer.__init__(self)
        self.translator_class = HTML5Translator
        return

    def translate(self):
        visitor = self.translator_class(self.document)
        self.document.walkabout(visitor)
        self.output = visitor.output
        self.head = visitor.head
        self.body = visitor.body
        self.title = visitor.title
        self.docinfo = visitor.docinfo
        return

    def assemble_parts(self):
        writers.Writer.assemble_parts(self)
        self.parts['head'] = self.head
        self.parts['body'] = self.body
        self.parts['title'] = self.title
        self.parts['docinfo'] = self.docinfo
        return

    def get_transforms(self):
        return writers.Writer.get_transforms(self) + [FooterToBottom]


class ElemStack(object):
    '''
    Helper class to handle nested contexts and indentation
    '''
    def __init__(self, settings):
        self.stack = []
        self.indent_level = 0
        self.indent_output = settings.indent_output
        self.indent_width = settings.tab_width

    def _indent_elem(self, element, indent):
        result = []
        # Indentation schema:
        #
        #         current position
        #                |
        #                v
        #           <tag>|
        # |   indent   |<elem>
        # |indent-1|</tag>
        #          ^
        #      ends here
        if self.indent_output and indent:
            indentation = '\n' + self.indent_width * self.indent_level * ' '
            result.append(indentation)
        result.append(element)
        if self.indent_output and indent:
            indentation = '\n' + self.indent_width * (self.indent_level - 1) * ' '
            result.append(indentation)
        return result

    def append(self, element, indent=True):
        '''
        Append to current element
        '''
        self.stack[-1].append(self._indent_elem(element, indent))
        return

    def begin_elem(self):
        '''
        Start a new element context
        '''
        self.stack.append([])
        self.indent_level += 1
        return

    def commit_elem(self, elem, indent=True):
        '''
        A new element is create by removing its stack to make a tag.
        This tag is pushed back into its parent's stack.
        '''
        pop = self.stack.pop()
        elem(*pop)
        self.indent_level -= 1
        self.append(elem, indent)
        return

    def pop(self):
        return self.pop_elements(1)[0]

    def pop_elements(self, num_elements):
        assert num_elements > 0
        parent_stack = self.stack[-1]
        result = []
        for x in range(num_elements):
            pop = parent_stack.pop()
            elem = pop[0 if len(pop) == 1 else self.indent_output]
            result.append(elem)
        result.reverse()
        return result


dv = 'default_visit'
dp = 'default_departure'
pass_ = 'no_op'


class HTML5Translator(nodes.NodeVisitor):

    rst_terms = {
        # 'term': ('tag', 'visit_func', 'depart_func', use_term_in_class,
        #          indent_elem)
        # use_term_in_class and indent_elem are optionals.
        # If not given, the default is False, True
        'Text': (None, 'visit_Text', None),
        'abbreviation': ('abbr', dv, dp),
        'acronym': (None, dv, dp),
        'address': (None, 'visit_address', None),
        'admonition': ('aside', 'visit_aside', 'depart_aside', True),
        'attention': ('aside', 'visit_aside', 'depart_aside', True),
        'attribution': ('p', dv, dp, True),
        'author': (None, 'visit_bibliographic_field', None),
        'authors': (None, 'visit_authors', None),
        'block_quote': ('blockquote', dv, dp),
        'bullet_list': ('ul', dv, dp, False),
        'caption': ('figcaption', dv, dp, False),
        'caution': ('aside', 'visit_aside', 'depart_aside', True),
        'citation': (None, 'visit_citation', 'depart_citation', True),
        'citation_reference': ('a', 'visit_citation_reference',
                               'depart_reference', True, False),
        'classifier': (None, 'visit_classifier', None),
        'colspec': (None, pass_, 'depart_colspec'),
        'comment': (None, 'skip_node', None),
        'compound': ('div', dv, dp, True),
        'contact': (None, 'visit_bibliographic_field', None),
        'container': ('div', dv, dp, True),
        'copyright': (None, 'visit_bibliographic_field', None),
        'danger': ('aside', 'visit_aside', 'depart_aside', True),
        'date': (None, 'visit_bibliographic_field', None),
        'decoration': (None, 'do_nothing', None),
        'definition': ('dd', dv, dp),
        'definition_list': ('dl', dv, dp),
        'definition_list_item': (None, 'do_nothing', None),
        'description': ('td', dv, dp),
        'docinfo': (None, 'do_nothing', None),
        'doctest_block': ('pre', 'visit_literal_block', 'depart_literal_block', True),
        'document': (None, 'visit_document', 'depart_document'),
        'emphasis': ('em', dv, dp, False, False),
        'entry': (None, dv, 'depart_entry'),
        'enumerated_list': ('ol', dv, 'depart_enumerated_list'),
        'error': ('aside', 'visit_aside', 'depart_aside', True),
        'field': (None, 'visit_field', None),
        'field_body': (None, 'do_nothing', None),
        'field_list': (None, 'do_nothing', None),
        'field_name': (None, 'do_nothing', None),
        'figure': (None, 'visit_figure', dp),
        'footer': (None, dv, dp),
        'footnote': (None, 'visit_citation', 'depart_citation', True),
        'footnote_reference': ('a', 'visit_citation_reference', 'depart_reference', True, False),
        'generated': (None, 'do_nothing', None),
        'header': (None, dv, dp),
        'hint': ('aside', 'visit_aside', 'depart_aside', True),
        'image': ('img', dv, dp),
        'important': ('aside', 'visit_aside', 'depart_aside', True),
        'inline': ('span', dv, dp, False, False),
        'label': ('th', 'visit_reference', 'depart_label'),
        'legend': ('div', dv, dp, True),
        'line': (None, 'visit_line', None),
        'line_block': ('pre', 'visit_line_block', 'depart_line_block', True),
        'list_item': ('li', dv, dp),
        'literal': ('code', 'visit_literal', 'depart_literal', False, False),
        'literal_block': ('pre', 'visit_literal_block', 'depart_literal_block'),
        'math': (None, 'visit_math_block', None),
        'math_block': (None, 'visit_math_block', None),
        'meta': (None, 'visit_meta', None),
        'note': ('aside', 'visit_aside', 'depart_aside', True),
        'option': ('kbd', 'visit_option', dp, False, False),
        'option_argument': ('var', 'visit_option_argument', dp, False, False),
        'option_group': ('td', 'visit_option_group', 'depart_option_group'),
        'option_list': (None, 'visit_option_list', 'depart_option_list', True),
        'option_list_item': ('tr', dv, dp),
        'option_string': (None, 'do_nothing', None),
        'organization': (None, 'visit_bibliographic_field', None),
        'paragraph': ('p', 'visit_paragraph', dp),
        'pending': (None, dv, dp),
        'problematic': ('a', 'visit_problematic', 'depart_reference', True, False),
        'raw': (None, 'visit_raw', None),
        'reference': ('a', 'visit_reference', 'depart_reference', False, False),
        'revision': (None, 'visit_bibliographic_field', None),
        'row': ('tr', 'visit_row', 'depart_row'),
        'rubric': ('p', dv, 'depart_rubric', True),
        'section': ('section', 'visit_section', 'depart_section'),
        'sidebar': ('aside', 'visit_aside', 'depart_aside', True),
        'status': (None, 'visit_bibliographic_field', None),
        'strong': (None, dv, dp, False, False),
        'subscript': ('sub', dv, dp, False, False),
        'substitution_definition': (None, 'skip_node', None),
        'substitution_reference': (None, 'skip_node', None),
        'subtitle': (None, 'visit_target', 'depart_subtitle'),
        'superscript': ('sup', dv, dp, False, False),
        'system_message': ('div', 'visit_system_message', dp),
        'table': (None, 'visit_table', 'depart_table'),
        'target': ('a', 'visit_target', 'depart_reference', False, False),
        'tbody': (None, dv, dp),
        'term': ('dt', dv, dp),
        'tgroup': (None, 'do_nothing', None),
        'thead': (None, 'visit_thead', 'depart_thead'),
        'tip': ('aside', 'visit_aside', 'depart_aside', True),
        'title': (None, dv, 'depart_title'),
        'title_reference': ('cite', dv, dp, False, False),
        'topic': ('aside', 'visit_aside', 'depart_aside', True),
        'transition': ('hr', dv, dp),
        'version': (None, 'visit_bibliographic_field', None),
        'warning': ('aside', 'visit_aside', 'depart_aside', True),
    }

    default_template = '<!DOCTYPE html>\n<html{html_attr}>\n' \
                       '<head>{head}</head>\n<body>{body}</body>\n</html>'

    def _map_terms_to_functions(self):
        '''
        Map terms to visit and departure functions
        '''
        for term, spec in self.rst_terms.items():
            visit_func = spec[1] and getattr(self, spec[1], self.unknown_visit)
            depart_func = spec[2] and getattr(self, spec[2], self.unknown_departure)
            if visit_func:
                setattr(self, 'visit_' + term, visit_func)
            if depart_func:
                setattr(self, 'depart_' + term, depart_func)
        return

    def _get_template(self, document):
        template = document.settings.template
        if not template:
            return self.default_template
        import os
        if os.path.isfile(template):
            from io import open
            with open(template, 'r', encoding='utf-8') as template_file:
                return template_file.read()
        else:
            return template

    def __init__(self, document):
        nodes.NodeVisitor.__init__(self, document)
        self.heading_level = int(getattr(self.document.settings, 'initial_header_level', 0))
        self.context = ElemStack(document.settings)
        self.template = self._get_template(document)
        self.docinfo = OrderedDict()
        self._parse_params()
        self._map_terms_to_functions()
        return

    def _parse_params(self):
        self.metatags = [tag.meta(charset=self.document.settings.output_encoding)]
        self.stylesheets = []
        stylesheets = self.document.settings.stylesheet or []
        for href in stylesheets:
            self.stylesheets.append(tag.link(rel='stylesheet', href=href))
        self.scripts = []
        scripts = self.document.settings.script or []
        for src, attributes in scripts:
            script = tag.script(src=src)
            if attributes:
                script = script(**{attributes: attributes})
            self.scripts.append(script)
        return

    def indent_head(self):
        if self.document.settings.indent_output:
            indent = '\n' + ' ' * self.document.settings.tab_width
            result = []
            for f in self.head:
                result.append(tag(indent, f))
            result.append('\n')
            self.head = result
        return

    def _get_template_values(self):
        html_attrs = self.document.settings.html_tag_attr
        html_attrs = html_attrs and ' ' + ' '.join(html_attrs) or ''
        self.head = self.metatags + self.stylesheets + self.scripts
        for key, value in self.docinfo.items():
            self.head.append(tag.meta(name=key, content=value))
        self.indent_head()
        self.head = ''.join(XHTMLSerializer()(tag(*self.head)))
        self.body = ''.join(XHTMLSerializer()(tag(*self.context.stack)))
        values = {}
        values['html_attr'] = html_attrs
        values['head'] = self.head
        values['body'] = self.body
        return values

    @property
    def output(self):
        values = self._get_template_values()
        return self.template.format(**values)

    def set_next_elem_attr(self, name, value):
        '''
        The given attribute will be inserted into the attributes of the
        next element.
        '''
        self.next_elem_attr = {name: value}

    def parse(self, node):
        '''
        Get tag name, indentantion and correct attributes of a node according
        to its class
        '''
        node_class_name = node.__class__.__name__
        spec = self.rst_terms[node_class_name]
        tag_name = spec[0] or node_class_name
        use_name_in_class = len(spec) > 3 and spec[3]
        indent = spec[4] if len(spec) > 4 else True
        if use_name_in_class:
            node['classes'].insert(0, node_class_name)

        attributes = {}
        replacements = {
            'refuri': 'href', 'uri': 'src', 'refid': 'href',
            'morerows': 'rowspan', 'morecols': 'colspan', 'classes': 'class',
            'ids': 'id',
        }
        ignores = (
            'names', 'dupnames', 'bullet', 'enumtype', 'colwidth', 'stub',
            'backrefs', 'auto', 'anonymous',
        )
        for k, v in node.attributes.items():
            if k in ignores or not v:
                continue
            if k in replacements:
                k = replacements[k]
            if isinstance(v, list):
                v = ' '.join(v)
            attributes[k] = v

        if getattr(self, 'next_elem_attr', None):
            attributes.update(self.next_elem_attr)
            del self.next_elem_attr
        return tag_name, indent, attributes

    def once_attr(self, name, default=None):
        '''
        The attribute is used once and then it is deleted
        '''
        if hasattr(self, name):
            result = getattr(self, name)
            delattr(self, name)
            return result
        else:
            return default

    def default_visit(self, node):
        '''
        Initiate a new context to store inner HTML5 elements.
        '''
        if 'ids' in node and self.once_attr('expand_id_to_anchor', default=True):
            # create an anchor <a id=id></a> for each id found before the
            # current element.
            for id in node['ids'][1:]:
                self.context.begin_elem()
                self.context.commit_elem(tag.a(id=id))
            node.attributes['ids'] = node.attributes['ids'][0:1]
        self.context.begin_elem()
        return

    def default_departure(self, node):
        '''
        Create the node's corresponding HTML5 element and combine it with its
        stored context.
        '''
        tag_name, indent, attributes = self.parse(node)
        elem = getattr(tag, tag_name)(**attributes)
        self.context.commit_elem(elem, indent)
        return

    def _compacted_paragraph(self, node):
        """
        Determine if the <p> tags around paragraph ``node`` can be omitted.
        Based on :func:`docutils.writers.html4css1.HTMLTranslator.should_be_compact_paragraph`
        """
        # extra oarenthesis for pep8 alignment conformity
        if ((isinstance(node.parent, (nodes.document, nodes.compound,
                        nodes.block_quote, nodes.system_message, )) or
             node['classes'] or 'paragraph' != node.__class__.__name__)):
            return False
        first = isinstance(node.parent[0], nodes.label)  # skip label
        for child in node.parent.children[first:]:
            # only first paragraph can be compact
            if isinstance(child, nodes.Invisible):
                continue
            if child is node:
                break
            return False
        parent_length = len([n for n in node.parent if not isinstance(
            n, (nodes.Invisible, nodes.label))])
        return parent_length == 1

    def visit_paragraph(self, node):
        if self._compacted_paragraph(node):
            raise nodes.SkipDeparture
        self.default_visit(node)

    def _strip_spaces(self, text):
        return re.sub(r'\s+', ' ', text)

    def visit_Text(self, node):
        text = node.astext()
        if not getattr(self, 'preserve_space', None):
            text = self._strip_spaces(text)
        self.context.append(text, indent=False)
        raise nodes.SkipDeparture

    def visit_section(self, node):
        self.heading_level += 1
        self.default_visit(node)

    def depart_section(self, node):
        self.heading_level -= 1
        self.default_departure(node)

    def depart_title(self, node):
        spec, indent, attr = self.parse(node)
        if isinstance(node.parent, nodes.table):
            elem = tag.caption(**attr)
        else:
            assert self.heading_level >= 0
            if self.heading_level == 0:
                self.heading_level = 1
            if 'href' in attr:
                # backref to toc entry
                # anchor = tag.a(href=("#" + attr['href']), class_="toc-backref")
                # self.context.commit_elem(anchor)
                # anchor = self.context.pop()
                # self.context.begin_elem()
                # self.context.append(anchor, indent=False)
                del attr['href']
            elem = getattr(tag, 'h' + unicode(self.heading_level))(**attr)
        self.context.commit_elem(elem, indent)

    def depart_subtitle(self, node):
        # mount the subtitle heading
        subheading_level = getattr(tag, 'h' + unicode(self.heading_level + 1))
        self.context.commit_elem(subheading_level)
        self.heading_level -= 1

    def depart_enumerated_list(self, node):
        '''
        Ordered list.
        It may have a preffix and suffix that must be handled by CSS3 and
        javascript to be presented as intended.

        .. seealso::

            * `Automatic numbering with CSS Counters <http://goo.gl/nXq02>`_
            * `How to add brackets to an ordered list? <http://goo.gl/Tdzmf>`_
        '''
        if 'enumtype' in node:
            enumtypes = {
                'arabic': '1',
                'loweralpha': 'a',
                'upperalpha': 'A',
                'lowerroman': 'i',
                'upperroman': 'I'
            }
            node['type'] = enumtypes.get(node['enumtype'], '1')
        if node.get('suffix') == '.' and 'preffix' not in node:
            # default suffix doesn't need special treatment
            del node['suffix']
        self.default_departure(node)

    def skip_node(self, node):
        """Internal only"""
        raise nodes.SkipNode

    def no_op(self, node):
        pass

    def do_nothing(self, node):
        '''
        equivalent to visit: pass and depart: pass
        '''
        raise nodes.SkipDeparture

    def visit_classifier(self, node):
        '''
        Classifier should remain beside the previous element
        '''
        term = self.context.pop()
        term(' ', tag.span(':', class_='classifier-delimiter'), ' ',
             tag.span(node.astext(), class_='classifier'))
        self.context.append(term)
        raise nodes.SkipNode

    #
    # table
    #

    def visit_table(self, node):
        self.th_required = 0
        self.in_thead = False
        self.default_visit(node)

    def depart_table(self, node):
        del self.th_required
        del self.in_thead
        self.default_departure(node)

    def depart_colspec(self, node):
        '''
        <col /> tags are not generated anymore because they're pretty useless
        since they cannot receive any attribute from a rst table.
        Anyway, there are better ways to apply styles to columns.
        See http://csswizardry.com/demos/zebra-striping/ for example.

        Nevertheless,
        if a colspec node with a "stub" attribute indicates that the column should be a th tag.
        '''
        if 'stub' in node:
            self.th_required += 1
        return

    def visit_thead(self, node):
        self.in_thead = True
        self.default_visit(node)

    def depart_thead(self, node):
        self.in_thead = False
        self.default_departure(node)

    def visit_row(self, node):
        self.th_available = self.th_required
        self.default_visit(node)

    def depart_row(self, node):
        del self.th_available
        self.default_departure(node)

    def depart_entry(self, node):
        if self.in_thead or self.th_available:
            name = 'th'
            self.th_available -= 1
        else:
            name = 'td'

        if 'morerows' in node:
            node['morerows'] = node['morerows'] + 1
        if 'morecols' in node:
            node['morecols'] = node['morecols'] + 1

        waste, indent, attr = self.parse(node)
        self.context.commit_elem(getattr(tag, name)(**attr))

    def visit_reference(self, node):
        if 'ids' in node:
            del node.attributes['ids']
        self.default_visit(node)
        return

    def depart_reference(self, node):
        if 'name' in node:
            del node['name']
        if 'refid' in node:
            node['refid'] = '#' + node['refid']
        self.default_departure(node)
        return

    def visit_target(self, node):
        if not node.astext():
            raise nodes.SkipNode
        self.expand_id_to_anchor = False
        self.default_visit(node)

    def visit_literal(self, node):
        self.preserve_space = getattr(self, 'preserve_space', 0) + 1
        self.default_visit(node)
        return

    def depart_literal(self, node):
        self.preserve_space -= 1
        if self.preserve_space == 0:
            del self.preserve_space
        if node['classes']:
            node['classes'] = node['classes'][-1]
        self.default_departure(node)
        return

    def visit_literal_block(self, node):
        if 'code' in node['classes']:
            self.context.begin_elem()  # <pre>. The next will be for <code>
        self.preserve_space = 1
        self.default_visit(node)
        return

    def depart_literal_block(self, node):
        del self.preserve_space
        if 'code' in node['classes']:
            code = node['classes'].index('code')
            node['classes'].pop(code)
            code = tag.code(class_=' '.join(node['classes']))
            self.context.commit_elem(code, indent=False)
            del node['classes']
        self.default_departure(node)
        return

    def visit_math_block(self, node):
        '''
        Only MathJax support
        '''
        math_code = node.astext()
        math_env = pick_math_environment(math_code)
        if 'align' in math_env:
            template = '\\begin{%s}\n%s\n\\end{%s}' % (math_env, math_code, math_env)
            elem = tag.div(template)
        else:  # equation
            template = '\(%s\)' % math_code
            elem = tag.span(template)
        elem(class_='math')
        self.context.append(elem)
        if not getattr(self, 'already_has_math_script', None):
            src = "http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML"
            self.scripts.append(tag.script(src=src))
            self.already_has_math_script = True
        raise nodes.SkipNode

    def visit_document(self, node):
        if 'title' in node:
            self.metatags.insert(0, tag.title(node['title']))
            self.title = node['title']
        else:
            self.title = ''
        self.expand_id_to_anchor = False
        self.default_visit(node)

    def depart_document(self, node):
        self.context.stack = self.context.stack[0]
        return

    def visit_raw(self, node):
        if 'html' in node.get('format', '').split():
            for line in node.astext().splitlines():
                self.context.append(Markup(line))
        raise nodes.SkipNode

    def visit_aside(self, node):
        self.save_heading_level = self.heading_level
        self.heading_level = 1
        self.default_visit(node)

    def depart_aside(self, node):
        self.heading_level = self.save_heading_level
        del self.save_heading_level
        if node['classes'] and node['classes'][0].startswith('admonition-'):
            del node['classes'][0]
        self.default_departure(node)

    def depart_rubric(self, node):
        node['classes'] = []
        self.default_departure(node)

    def visit_field(self, node):
        self.docinfo[node.children[0].astext()] = self._strip_spaces(node.children[1].astext())
        raise nodes.SkipNode

    def visit_bibliographic_field(self, node):
        self.docinfo[node.__class__.__name__] = self._strip_spaces(node.astext())
        raise nodes.SkipNode

    def visit_address(self, node):
        self.docinfo[node.__class__.__name__] = ', '.join(node.astext().split('\n'))
        raise nodes.SkipNode

    def visit_authors(self, node):
        self.docinfo[node.__class__.__name__] = '; '.join(node.astext().split())
        raise nodes.SkipNode

    def visit_option_list(self, node):
        self.context.begin_elem()  # table
        self.context.begin_elem()  # tbody

    def depart_option_list(self, node):
        self.context.commit_elem(tag.tbody)
        waste, waste_, attr = self.parse(node)
        self.context.commit_elem(tag.table(**attr))

    def visit_citation(self, node):
        self.visit_option_list(node)
        self.context.begin_elem()  # tr

    def depart_citation(self, node):
        # td initiated at depart_label
        self.context.commit_elem(tag.td)
        self.context.commit_elem(tag.tr)
        self.depart_option_list(node)

    def visit_option_group(self, node):
        self.option_level = 0
        self.default_visit(node)

    def depart_option_group(self, node):
        if self.document.settings.option_limit and \
                len(node.astext()) > self.document.settings.option_limit:
            node['morecols'] = 2
            self.default_departure(node)
            self.context.commit_elem(tag.tr)  # closes this tr
            self.context.begin_elem()  # begins another tr
            self.context.append(tag.td)  # empty td due to colspan
        else:
            self.default_departure(node)

    def visit_option(self, node):
        if self.option_level:
            self.context.append(', ', indent=False)
        self.option_level += 1
        self.default_visit(node)

    def visit_option_argument(self, node):
        if 'delimiter' in node:
            self.context.append(node['delimiter'], indent=False)
            del node.attributes['delimiter']
        self.default_visit(node)

    def visit_citation_reference(self, node):
        """
        Instead of a typical visit_reference call this def is required to
        remove the backref id that is included but not used in rst2html5.
        """
        self.expand_id_to_anchor = False
        self.default_visit(node)

    def depart_label(self, node):
        self.default_departure(node)
        self.context.begin_elem()  # next td

    def visit_line(self, node):
        self.line_level = getattr(self, 'line_level', -1) + 1
        if self.line_level:
            tab_width = self.document.settings.tab_width
            separator = '\n' + ' ' * tab_width * (self.line_block_level - 1)
            self.context.append(separator, indent=False)
        raise nodes.SkipDeparture

    def visit_line_block(self, node):
        '''
        Line blocks use <pre>.
        Lines breaks and spacing are reconstructured based on line_block_level
        '''
        self.line_block_level = getattr(self, 'line_block_level', 0) + 1
        if self.line_block_level == 1:
            self.default_visit(node)

    def depart_line_block(self, node):
        self.line_block_level -= 1
        if self.line_block_level == 0:
            del self.line_block_level
            del self.line_level
            self.default_departure(node)

    def visit_meta(self, node):
        waste, waste_, attr = self.parse(node)
        self.metatags.append(tag.meta(**attr))
        raise nodes.SkipNode

    def visit_problematic(self, node):
        self.expand_id_to_anchor = False
        if len(node['ids']) > 1:
            node['ids'] = node['ids'][0]
        self.default_visit(node)
        return

    def visit_system_message(self, node):
        self.default_visit(node)
        self.context.begin_elem()  # h1
        backrefs = [tag(' ', tag.a(v, href="#" + v)) for v in node['backrefs']]
        node.attributes.setdefault('line', '')
        text = 'System Message: {type}/{level} ({source} line ' \
               '{line})'.format(**node.attributes)
        h1 = tag.h1(text, *backrefs)
        self.context.commit_elem(h1)
        node.attributes = {'classes': [], 'ids': node.attributes['ids']}
        return

    def visit_figure(self, node):
        # move up the ids of child img
        for child in node:
            if isinstance(child, nodes.image) and 'ids' in child:
                node['ids'].extend(child['ids'])
                child['ids'] = []

        self.default_visit(node)


def main():
    from docutils.core import publish_cmdline, default_description

    description = 'Generates (X)HTML5 documents from standalone ' \
                  'reStructuredText sources.' + default_description
    publish_cmdline(writer=HTML5Writer(), description=description)

if __name__ == '__main__':
    main()




def settings(request):
    """inject few waliki's settings to the context  to be used in templates"""
    from waliki.settings import WALIKI_USE_MATHJAX      # NOQA
    return {k: v for (k, v) in locals().items() if k.startswith('WALIKI')}



from django.conf.urls import patterns, url
from waliki.settings import WALIKI_SLUG_PATTERN


urlpatterns = patterns('waliki.git.views',

    url(r'^_whatchanged/(?P<pag>\d+)$', 'whatchanged', name='waliki_whatchanged'),       # noqa
    url(r'^_whatchanged$', 'whatchanged', {'pag': '1'}, name='waliki_whatchanged'),       # noqa

    url(r'^_hooks/pull/(?P<remote>[a-zA-Z0-9]+)$', 'webhook_pull', name='waliki_webhook_pull'),
    url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/history/(?P<pag>\d+)$', 'history', name='waliki_history'),
    url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/history/$', 'history', {'pag': '1'}, name='waliki_history'),

    url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/version/(?P<version>[0-9a-f\^]{4,40})/raw$', 'version', {'raw': True},
        name='waliki_version_raw'),
    url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/version/(?P<version>[0-9a-f\^]{4,40})$', 'version', name='waliki_version'),

    url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/diff/(?P<old>[0-9a-f\^]{4,40})\.\.(?P<new>[0-9a-f\^]{4,40})/raw$',
        'diff', {'raw': True}, name='waliki_diff_raw'),

    url(r'^(?P<slug>' + WALIKI_SLUG_PATTERN + ')/diff/(?P<old>[0-9a-f\^]{4,40})\.\.(?P<new>[0-9a-f\^]{4,40})$', 'diff', name='waliki_diff'),
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import waliki.settings


class Migration(migrations.Migration):

    dependencies = [
        ('waliki', '0003_auto_20141110_0052'),
    ]

    operations = [
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(serialize=False, verbose_name='ID', primary_key=True, auto_created=True)),
                ('file', models.FileField(upload_to=waliki.settings.WALIKI_UPLOAD_TO)),
                ('page', models.ForeignKey(related_name='attachments', to='waliki.Page')),
            ],
            options={
                'verbose_name': 'Attachment',
                'verbose_name_plural': 'Attachments',
            },
            bases=(models.Model,),
        ),
    ]



import os
import re
import mimetypes
import unicodedata
from django.http import HttpResponse
from django.core.urlresolvers import reverse
from django.utils.six import PY2
from django.utils.encoding import force_text
from django.utils.safestring import mark_safe

def get_slug(text):
    def slugify(value):
        """
        same than django slugify but allowing uppercase and underscore
        """
        value = force_text(value)
        value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
        value = re.sub('[^\w\s\/_-]', '', value).strip()
        return mark_safe(re.sub('[-\s]+', '-', value))

    if PY2:
        from django.utils.encoding import force_unicode
        text = force_unicode(text)
    for sep in ('_', '/'):
        text = sep.join(slugify(t) for t in text.split(sep))
    return text.strip('/')


def get_url(text, *args):
    # *args needed to receive prefix and suffix for markdowns wikilinks ext
    from waliki.settings import get_slug
    slug = get_slug(text)
    if slug:
        return reverse('waliki_detail', args=(get_slug(text),))
    return ''


def send_file(path, filename=None, content_type=None):
    # TODO : remove it and use django-sendfile instead
    if filename is None:
        filename = os.path.basename(path)
    if content_type is None:
        content_type, encoding = mimetypes.guess_type(filename)
    response = HttpResponse(content_type=content_type)
    response['Content-Disposition'] = 'attachment; filename=%s' % filename
    response.write(open(path, "rb").read())
    return response



import re
import sys
from markups import (MarkdownMarkup as MarkdownMarkupBase,
                     TextileMarkup as TextileMarkupBase,
                     ReStructuredTextMarkup as ReStructuredTextMarkupBase)
from pyquery import PyQuery
from .utils import get_url
from waliki.settings import WALIKI_AVAILABLE_MARKUPS


class MarkdownMarkup(MarkdownMarkupBase):
    codemirror_mode_name = codemirror_mode = 'markdown'
    IMAGE_LINE = '![](%(url)s)'
    LINK_LINE = '[%(filename)s](<%(url)s>)'


    def __init__(self, filename=None, extensions=None, extension_configs=None):
        super(MarkdownMarkup, self).__init__(filename)
        self.md.set_output_format('html5')
        self.md.registerExtensions(extensions, extension_configs)


class ReStructuredTextMarkup(ReStructuredTextMarkupBase):

    IMAGE_LINE = '.. image:: %(url)s'
    LINK_LINE = '`%(filename)s <%(url)s>`_'

    codemirror_mode = codemirror_mode_name = 'rst'

    def __init__(self, filename=None, **kwargs):
        settings_overrides = kwargs.pop('settings_overrides', None)
        self.reader = kwargs.pop('reader', None)
        super(ReStructuredTextMarkup, self).__init__(filename, settings_overrides)
        if not self.reader:
            from waliki.directives.transforms import WalikiReader
            self.reader = WalikiReader()

        self.kwargs = kwargs

    def publish_parts(self, text):
        if 'rest_parts' in self._cache:
            return self._cache['rest_parts']
        parts = self._publish_parts(text, source_path=self.filename,
                                    settings_overrides=self.overrides,
                                    reader=self.reader, **self.kwargs)
        if self._enable_cache:
            self._cache['rest_parts'] = parts
        return parts

    def get_document_body(self, text):
        html = super(ReStructuredTextMarkup, self).get_document_body(text)
        if not html:
            return html

        # Convert unknown links to internal wiki links.
        # Examples:
        #   Something_ will link to '/something'
        #  `something great`_  to '/something-great'
        #  `another thing <thing>`_  '/thing'
        refs = [a.text[:-1] for a in PyQuery(html)('a.problematic') if not re.match(r'\|(.*)\|', a.text)]
        # substitions =  [a.text[:-1] for a in PyQuery(html)('a.problematic') if re.match(r'\|(.*)\|', a.text)]
        if refs:
            refs = '\n'.join('.. _%s: %s' % (ref, get_url(ref))
                             for ref in refs if get_url(ref))
            html = super(ReStructuredTextMarkup, self).get_document_body(
                        text + '\n\n' + refs)
        return html


class TextileMarkup(TextileMarkupBase):
    codemirror_mode_name = codemirror_mode = 'textile'


def get_all_markups():
    return [find_markup_class_by_name(s) for s in WALIKI_AVAILABLE_MARKUPS]


def find_markup_class_by_name(name):
    for markup in (ReStructuredTextMarkup, MarkdownMarkup, TextileMarkup):
        if markup.name.lower() == name.lower():
            return markup


def find_markup_class_by_extension(extension):
    for markup in (ReStructuredTextMarkup, MarkdownMarkup, TextileMarkup):
        if extension.lower() in markup.file_extensions:
            return markup



import os.path
from importlib import import_module
import collections
from django.conf import settings
from .utils import get_url
from waliki.rst2html5 import HTML5Writer
from waliki.plugins import str2object


def _get_default_data_dir(directory, abspath=True):
    settings_mod = import_module(settings.SETTINGS_MODULE)

    project_dir = os.path.dirname(settings_mod.__name__)
    if abspath:
        project_dir = os.path.abspath(project_dir)
    return os.path.join(project_dir, directory)


def deep_update(d, u):
    """update nested dicts. if u is a dict and hasn't a key, the original is keeped
    inspired in Alex Martelli's
    http://stackoverflow.com/a/3233356"""
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            r = deep_update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def _get_markup_settings(user_settings):
    defaults = {'reStructuredText': {
                    # check http://docutils.sourceforge.net/docs/user/config.html
                    'settings_overrides': {              # noqa
                        'initial_header_level': 2,
                        'record_dependencies': True,
                        'stylesheet_path': None,
                        'link_stylesheet': True,
                        'syntax_highlight': 'short',
                        'halt_level': 5,
                    },
                    'writer': HTML5Writer(),
                    'writer_name': 'html5',
                    },
                'Markdown': {
                    'extensions': ['wikilinks', 'headerid'],
                    'extension_configs': {
                        'wikilinks': {'build_url': get_url},
                        'headerid': {'level': 2},
                    }
                }
            }

    deep_update(defaults, user_settings)
    return defaults

WALIKI_AVAILABLE_MARKUPS = getattr(settings, 'WALIKI_AVAILABLE_MARKUPS', ['reStructuredText', 'Markdown'])

# options: reStructuredText, Markdown, Textile
WALIKI_DEFAULT_MARKUP = WALIKI_AVAILABLE_MARKUPS[0]


WALIKI_SLUG_PATTERN = getattr(settings, 'WALIKI_SLUG_PATTERN', '[a-zA-Z0-9-_\/]+')

WALIKI_SLUGIFY_FUNCTION = getattr(settings, 'WALIKI_SLUGIFY_FUNCTION', 'waliki.utils.get_slug')

get_slug = str2object(WALIKI_SLUGIFY_FUNCTION)


# your content folder. by default it's <project_root>/waliki_data
WALIKI_DATA_DIR = getattr(settings, 'WALIKI_DATA_DIR', None) or _get_default_data_dir('waliki_data')

# wich page is shown as the wiki index?
WALIKI_INDEX_SLUG = getattr(settings, 'WALIKI_INDEX_SLUG', "home")

# kwargs for each markup.
WALIKI_MARKUPS_SETTINGS = _get_markup_settings(getattr(settings, 'WALIKI_MARKUPS_SETTINGS', {}))

# get as txt
WALIKI_PDF_INCLUDE_TITLE = getattr(settings, 'WALIKI_PDF_INCLUDE_TITLE', False)

WALIKI_CODEMIRROR_SETTINGS = getattr(settings, 'WALIKI_CODEMIRROR_SETTINGS',
                                     {'lineNumbers': False, 'theme': 'monokai', 'autofocus': True})

# ('view_page', 'add_page', 'change_page', 'delete_page')
WALIKI_ANONYMOUS_USER_PERMISSIONS = getattr(settings, 'WALIKI_ANONYMOUS_USER_PERMISSIONS',
                                            ('view_page',))

WALIKI_LOGGED_USER_PERMISSIONS = getattr(settings, 'WALIKI_LOGGED_USER_PERMISSIONS',
                                         ('view_page', 'add_page', 'change_page'))

WALIKI_RENDER_403 = getattr(settings, 'WALIKI_RENDER_403', True)

WALIKI_PAGINATE_BY = getattr(settings, 'WALIKI_PAGINATE_BY', 20)

WALIKI_COMMITTER_EMAIL = getattr(settings, 'WALIKI_COMMITTER_EMAIL', 'waliki@waliki.pythonanywhere.com')

WALIKI_COMMITTER_NAME = getattr(settings, 'WALIKI_COMMITTER_NAME', 'Waliki')

WALIKI_CACHE_TIMEOUT = getattr(settings, 'WALIKI_CACHE_TIMEOUT', 60*60*24)

WALIKI_ATTACHMENTS_DIR = getattr(settings, 'WALIKI_ATTACHMENTS_DIR', None)  or _get_default_data_dir('waliki_attachments', False)

WALIKI_UPLOAD_TO_PATTERN = '%(slug)s/%(filename)s'

WALIKI_RST_DIRECTIVES = getattr(settings, 'WALIKI_RST_DIRECTIVES', ['waliki.directives.embed'])

WALIKI_RST_TRANSFORMS = getattr(settings, 'WALIKI_RST_TRANSFORMS', ['waliki.directives.transforms.Emojis'])

WALIKI_USE_MATHJAX = getattr(settings, 'WALIKI_USE_MATHJAX', False)


def WALIKI_UPLOAD_TO(instance, filename):
    return os.path.join(WALIKI_ATTACHMENTS_DIR,
                        WALIKI_UPLOAD_TO_PATTERN % {'slug': instance.page.slug,
                                                       'page_id': getattr(instance.page, 'id', ''),
                                                       'filename': filename,
                                                       'filename_extension': os.path.splitext(filename)[1]})

for mod in WALIKI_RST_DIRECTIVES:
    register_directive = str2object(mod + '.register_directive')
    register_directive()


from django import forms
from django.utils.translation import ugettext_lazy as _
from .models import Page
from ._markups import get_all_markups
from .settings import WALIKI_CODEMIRROR_SETTINGS as CM_SETTINGS, get_slug
from .acl import check_perms


class DeleteForm(forms.Form):
    what = forms.ChoiceField(label=_('What do you want to delete?'),
                             choices=(('this', _('Just this page')),
                                      ('namespace', _('This page and all the namespace')))
                             )


class MovePageForm(forms.ModelForm):
    slug = forms.CharField()
    just_redirect = forms.BooleanField(label=_('Just create a redirection'), widget=forms.HiddenInput, initial=False, required=False)

    class Meta:
        model = Page
        fields = ('slug', 'just_redirect',)
        exclude = ('slug',)

    def clean(self):
        cleaned_data = super(MovePageForm, self).clean()
        slug = cleaned_data['slug']
        just_redirect = cleaned_data.get('just_redirect', False)
        if self.instance.slug == slug:
            raise forms.ValidationError(_("The slug wasn't changed"))

        if Page.objects.filter(slug=slug).exists() and not just_redirect:
            self.fields['just_redirect'].widget = forms.CheckboxInput()
            raise forms.ValidationError(_("There is already a page with this slug"))

        return cleaned_data


class NewPageForm(forms.ModelForm):

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop('user', None)
        super(NewPageForm, self).__init__(*args, **kwargs)

    class Meta:
        model = Page
        fields = ['title', 'slug', 'markup']

    def clean_slug(self):
        slug = self.cleaned_data['slug']
        if not slug:
            raise forms.ValidationError(_("The slug can't be empty"))
        if get_slug(slug) != slug:
            raise forms.ValidationError(_("The slug isn't valid"))
        if not check_perms(['add_page'], self.user, slug):
            raise forms.ValidationError(_("You have no permission to create a page with this slug"))
        if Page.objects.filter(slug=slug).exists():
            raise forms.ValidationError(_("There is already a page with this slug"))

        return slug



class PageForm(forms.ModelForm):
    raw = forms.CharField(label="", widget=forms.Textarea)
    # Translators: log message
    message = forms.CharField(label=_('Log message'), max_length=200, required=False)
    extra_data = forms.CharField(widget=forms.HiddenInput, required=False)

    class Meta:
        model = Page
        fields = ['title', 'markup', 'raw', 'message']

    class Media:
        modes = tuple('codemirror/mode/%s/%s.js' % (m.codemirror_mode, m.codemirror_mode)
                      for m in get_all_markups() if hasattr(m, 'codemirror_mode'))
        theme = ('codemirror/theme/%s.css' % CM_SETTINGS['theme'],) if 'theme' in CM_SETTINGS else ()

        js = ('codemirror/lib/codemirror.js',
              'codemirror/addon/mode/overlay.js') + modes + ('js/waliki_editor.js',)
        css = {
            'all': ('codemirror/lib/codemirror.css', ) + theme
        }

    def __init__(self, *args, **kwargs):
        is_hidden = kwargs.pop('is_hidden', None)
        super(PageForm, self).__init__(*args, **kwargs)
        self.fields['raw'].initial = self.instance.raw
        self.fields['markup'].widget = forms.HiddenInput()
        self.fields['message'].widget = forms.TextInput(attrs={'placeholder': _('Update %s') % self.instance.path})
        if is_hidden:
            for field in self.fields.values():
                field.widget = forms.HiddenInput()

    def save(self, commit=True):
        instance = super(PageForm, self).save(commit)
        if commit:
            instance.raw = self.cleaned_data['raw']

        return instance



from django.utils.translation import ugettext_lazy as _
from waliki.plugins import BasePlugin, register


class PdfPlugin(BasePlugin):
    slug = 'pdf'
    urls_page = ['waliki.pdf.urls']
    extra_page_actions = {'reStructuredText': [('waliki_pdf', _('Get as PDF'))]}

register(PdfPlugin)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('waliki', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='aclrule',
            name='apply_to',
            field=models.CharField(choices=[('any', 'Any user'), ('logged', 'Any logged in user'), ('staff', 'Any staff member'), ('superusers', 'Any superuser'), ('explicit', 'Any user/group explicitly defined')], max_length=25, default='explicit'),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import waliki.settings


class Migration(migrations.Migration):

    dependencies = [
        ('attachments', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attachment',
            name='file',
            field=models.FileField(upload_to=waliki.settings.WALIKI_UPLOAD_TO, max_length=300),
        ),
    ]



import tempfile
from sh import rst2pdf
from django.shortcuts import get_object_or_404
from waliki.models import Page
from waliki.utils import send_file
from waliki.settings import WALIKI_PDF_INCLUDE_TITLE
from waliki.acl import permission_required


@permission_required('view_page')
def pdf(request, slug):
    page = get_object_or_404(Page, slug=slug)

    with tempfile.NamedTemporaryFile(suffix='.pdf') as output:
        outfile = output.name
    if WALIKI_PDF_INCLUDE_TITLE:
        line = "/" * len(page.title)
        title = "%s\n%s\n%s\n\n" % (line, page.title, line)
        with tempfile.NamedTemporaryFile(suffix='.rst', mode="w", delete=False) as infile:
            infile.file.write(title + page.raw)
            infile = infile.name
    else:
        infile = page.abspath
    rst2pdf(infile, o=outfile)
    filename = page.title.replace('/', '-').replace('..', '-')
    return send_file(outfile, filename="%s.pdf" % filename)



from django import template
from django.core.urlresolvers import reverse
from waliki.acl import check_perms as check_perms_helper
from waliki.models import Page
from waliki.forms import PageForm


register = template.Library()


@register.inclusion_tag('waliki/extra_page_actions.html')
def extra_page_actions(page):
    from waliki.plugins import get_extra_page_actions
    return {'page': page, 'extra_page_actions': get_extra_page_actions()}


@register.inclusion_tag('waliki/extra_edit_actions.html')
def extra_edit_actions(page):
    from waliki.plugins import get_extra_edit_actions
    return {'page': page, 'extra_edit_actions': get_extra_edit_actions()}


@register.inclusion_tag('waliki/entry_point.html', takes_context=True)
def entry_point(context, block_name):
    """include an snippet at the bottom of a block, if it exists

    For example, if the plugin with slug 'attachments' is registered

       waliki/attachments_edit_content.html  will be included with

        {% entry_point 'edit_content' %}

    which is declared at the bottom of the block 'content' in edit.html
    """
    from waliki.plugins import get_plugins
    includes = []
    for plugin in get_plugins():
        template_name = 'waliki/%s_%s.html' % (plugin.slug, block_name)
        try:
            # template exists
            template.loader.get_template(template_name)
            includes.append(template_name)
        except template.TemplateDoesNotExist:
            continue
    context.update({'includes': includes})
    return context

@register.inclusion_tag('waliki/navbar_links.html')
def navbar_links():
    from waliki.plugins import get_navbar_links
    return {'navbar_links': get_navbar_links()}


@register.filter(name="getattr")
def get_attr(obj, val):
    try:
        return getattr(obj, val)
    except AttributeError:
        try:
            return obj[val]
        except (KeyError, TypeError):
            return None


class CheckPermissionsNode(template.Node):

    def __init__(self, perms, user, slug, context_var):
        self.perms = template.Variable(perms)
        self.user = template.Variable(user)
        self.slug = template.Variable(slug)
        self.context_var = context_var

    def render(self, context):
        perms = [perm.strip() for perm in self.perms.literal.split(',')]
        user = self.user.resolve(context)
        slug = self.slug.literal or self.slug.resolve(context)
        if slug:
            context[self.context_var] = check_perms_helper(perms, user, slug)
        else:
            context[self.context_var] = False
        return ''


@register.tag
def check_perms(parser, token):
    """
    Returns a list of permissions (as ``codename`` strings) for a given
    ``user``/``group`` and ``obj`` (Model instance).

    Parses ``check_perms`` tag which should be in format::

        {% check_perms "perm1[, perm2, ...]" for user in slug as "context_var" %}

    or

        {% check_perms "perm1[, perm2, ...]" for user in "slug" as "context_var" %}

    .. note::

       Make sure that you set and use those permissions in same template
       block (``{% block %}``).

    Example of usage (assuming ``page` objects are available from *context*)::

        {% check_perms "delete_page" for request.user in page.slug as "can_delete" %}
        {% if can_delete %}
            ...
        {% endif %}

    """
    bits = token.split_contents()
    format = '{% check_perms "perm1[, perm2, ...]" for user in slug as "context_var" %}'
    if len(bits) != 8 or bits[2] != 'for' or bits[4] != "in" or bits[6] != 'as':
        raise template.TemplateSyntaxError("get_obj_perms tag should be in "
                                           "format: %s" % format)
    perms = bits[1]
    user = bits[3]
    slug = bits[5]
    context_var = bits[7]
    if perms[0] != perms[-1] or perms[0] not in ('"', "'"):
        raise template.TemplateSyntaxError("check_perms tag's perms "
                                           "argument should be in quotes")

    if context_var[0] != context_var[-1] or context_var[0] not in ('"', "'"):
        raise template.TemplateSyntaxError("check_perms tag's context_var "
                                           "argument should be in quotes")
    context_var = context_var[1:-1]
    return CheckPermissionsNode(perms, user, slug, context_var)


@register.inclusion_tag("waliki/box.html", takes_context=True)
def waliki_box(context, slug, show_edit=True, *args, **kwargs):
    """
    A templatetag to render a wiki page content as a box in any webpage,
    and allow rapid edition if you have permission.

    It's inspired in `django-boxes`_

    .. _django-boxes: https://github.com/eldarion/django-boxes
    """

    request = context["request"]
    try:
        page = Page.objects.get(slug=slug)
    except Page.DoesNotExist:
        page = None

    if (page and check_perms_helper('change_page', request.user, slug)
            or (not page and check_perms_helper('add_page', request.user, slug))):
        form = PageForm(instance=page, initial={'slug': slug})
        form_action = reverse("waliki_edit", args=[slug])
    else:
        form = None
        form_action = None

    return {
        "request": request,
        "slug": slug,
        "label": slug.replace('/', '_'),
        "page": page,
        "form": form,
        "form_action": form_action,
    }
